"""Dialog enabling text replacement."""

from qtpy.QtWidgets import QLayout, QWidget

import qtextra.helpers as hp
from qtextra.utils.table_config import TableConfig
from qtextra.widgets.qt_button_tag import QtTagButton
from qtextra.widgets.qt_dialog import QtDialog
from qtextra.widgets.qt_layout_scroll import QtScrollableHLayoutWidget
from qtextra.widgets.qt_table_view_check import QtCheckableTableView

SEPARATOR = " → "


class QtTextReplace(QtDialog):
    """Text replace dialog."""

    TABLE_CONFIG = (
        TableConfig()  # type: ignore
        .add("original text", "old", "int", 300, sizing="stretch")
        .add("new text", "new", "str", 300, sizing="stretch")
        # .add("changed", "changed", "str", 50, sizing="stretch")
    )

    def __init__(
        self,
        parent: QWidget,
        texts: list[str],
        message: str = "",
    ):
        self.message = message
        self.original_texts = texts
        self.search_and_replace_map: dict[str, str] = {}
        self.search_for_text = ""
        self.replace_with_text = ""
        super().__init__(parent)
        self.setMinimumWidth(600)
        self.search_edit.setFocus()
        self.populate()

    def populate(self) -> None:
        """Populate table."""
        for text in self.original_texts:
            self.table.add_row([text, text])

    def on_preview(self) -> None:
        """Preview text."""
        texts = self.table.get_col_data(self.TABLE_CONFIG.old)
        self.search_for_text = self.search_edit.text()
        self.replace_with_text = self.replace_edit.text()
        for row, current in enumerate(texts):
            changed, new_text = self.process_text(current)
            self.table.set_value(self.TABLE_CONFIG.new, row, new_text)
            # self.table.set_value(self.TABLE_CONFIG.changed, row, "True" if changed else "False")

    def on_add(self) -> None:
        """Add replacement text."""
        search_for = self.search_edit.text()
        replace_with = self.replace_edit.text()
        self.add(search_for, replace_with)

    def add(self, search_for: str, replace_with: str) -> None:
        """Add text replacement."""
        if search_for in self.search_and_replace_map and self.search_and_replace_map[search_for] == replace_with:
            return
        if search_for == replace_with:
            return

        self.search_and_replace_map[search_for] = replace_with
        self.search_edit.setText("")
        self.replace_edit.setText("")

        button = QtTagButton(
            f"{search_for}{SEPARATOR}{replace_with}",
            f"{search_for}{SEPARATOR}{replace_with}",
            allow_selected=False,
            action_type="delete",
            action_icon="cross",
        )
        button.evt_action.connect(self.on_remove)
        self.search_replace_layout.insertWidget(0, button)
        self.on_preview()

    def on_remove(self, hash_id: str) -> None:
        """Remove filter."""
        hash_ids = [
            self.search_replace_layout.get_widget(index).tag for index in range(self.search_replace_layout.count() - 1)
        ]
        index = hash_ids.index(hash_id)
        search_for, replace_with = hash_id.split(SEPARATOR)
        del self.search_and_replace_map[search_for]
        self.search_replace_layout.removeWidgetOrLayout(index)
        self.on_preview()

    def process_text(self, current: str) -> tuple[bool, str]:
        """Process single text item."""
        new_text = current
        for search, replace in self.search_and_replace_map.items():
            new_text = new_text.replace(search, replace)
        new_text = new_text.replace(self.search_for_text, self.replace_with_text)
        return new_text != current, new_text

    @property
    def new_texts(self) -> list[str]:
        """Text that was entered."""
        new_texts = []
        for current in self.original_texts:
            _, new_text = self.process_text(current)
            new_texts.append(new_text)
        return new_texts

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QLayout:
        """Make panel."""
        self.table = QtCheckableTableView(self, config=self.TABLE_CONFIG, sortable=True, checkable=False)
        self.table.setCornerButtonEnabled(False)
        hp.set_font(self.table, 10)
        self.table.setup_model(
            self.TABLE_CONFIG.header, self.TABLE_CONFIG.no_sort_columns, self.TABLE_CONFIG.hidden_columns
        )

        self.search_edit = hp.make_line_edit(
            self,
            "",
            tooltip="Text that will be searched for in the document to replace.",
            placeholder="Text to search for...",
            func_changed=self.on_preview,
        )
        self.replace_edit = hp.make_line_edit(
            self,
            "",
            tooltip="Text that will be used to replace the searched text.",
            placeholder="Text t o replace with...",
            func_changed=self.on_preview,
        )

        self.search_replace_layout = QtScrollableHLayoutWidget(self)
        self.search_replace_layout.set_min_height(self.search_edit.height())

        layout = hp.make_form_layout()
        layout.addRow("Search text:", self.search_edit)
        layout.addRow("Replace text:", self.replace_edit)
        layout.addRow(hp.make_btn(self, "Add", func=self.on_add))
        layout.addRow(self.search_replace_layout)
        layout.addRow(self.table)
        layout.addRow(
            hp.make_h_layout(
                hp.make_btn(self, "OK", func=self.accept),
                hp.make_btn(self, "Cancel", func=self.reject),
            )
        )
        return layout


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import apply_style, qapplication

        _ = qapplication()  # analysis:ignore
        dlg = QtTextReplace(
            None,
            [
                "This is a test.",
                "This is a test test.",
                "This is a test test test.",
                "Replace me with something else.",
                "Replace me with something else.",
            ],
        )
        apply_style(dlg)
        dlg.show()
        sys.exit(dlg.exec_())  # type: ignore[attr-defined]

    _main()  # type: ignore[no-untyped-call]



"""Slightly modified QToolButton with nicer interface."""

from typing import Callable, List, Optional, Tuple, Union

from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QCursor, QIcon
from qtpy.QtWidgets import QMenu, QToolButton


class QtToolButton(QToolButton):
    """QToolButton."""

    _icon, _pixmap = None, None

    def __init__(self, parent, text="", icon: Union[QIcon, None] = None):
        super().__init__(parent)
        self.setText(text)
        self.setPopupMode(self.ToolButtonPopupMode.MenuButtonPopup)
        self.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

        # Widget setup
        self._icon_size = QSize(16, 16)
        if icon:
            self.set_icon(icon)

    def set_menu(self, menu: QMenu, action: Optional[Callable] = None, auto_pop: bool = True) -> None:
        """Set tool button menu."""
        if not isinstance(menu, QMenu):
            raise ValueError("'menu' argument must be a QMenu instance")

        self.setMenu(menu)
        if action:
            if not callable(action):
                raise ValueError("'action' argument must be a callable")
            menu.triggered.connect(action)
        if auto_pop:
            self.pressed.connect(self.autopop)

    def set_icon(self, icon: Union[QIcon, None]) -> None:
        """Set the icon for the status bar widget."""
        if icon is not None and isinstance(icon, QIcon):
            self._icon = icon
            self._pixmap = icon.pixmap(self._icon_size)
            self.setIcon(icon)
            self.setMinimumWidth(37)  # based on an icon of size (16, 16)

    def set_size(self, size: Union[List, Tuple]) -> None:
        """Set size of the button."""
        if isinstance(size, (list, tuple)):
            self.setMinimumSize(QSize(*size))
            self.setMaximumSize(QSize(*size))

    def autopop(self) -> None:
        """When button is clicked (not necessarily on the down arrow, menu will be shown."""
        self.menu().popup(QCursor.pos())

    def popup(self) -> None:
        """When button is clicked (not necessarily on the down arrow, menu will be shown."""
        self.menu().popup(QCursor.pos())


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe()

        menu = QMenu(None)
        menu.addAction("Action 1")

        btn1 = QtToolButton(frame, text="None")
        btn1.set_menu(menu, print)

        ha.addWidget(btn1)
        frame.show()
        sys.exit(app.exec_())

    _main()  # type: ignore[no-untyped-call]



import io
import traceback
import typing as ty


def format_exception(exc_info: ty.Union[ty.Tuple, ty.List, Exception]):
    """Format exception."""
    if isinstance(exc_info, BaseException):
        exc_info = (type(exc_info), exc_info, exc_info.__traceback__)

    if not isinstance(exc_info, (list, tuple)):
        raise ValueError("Cannot format exception")

    # if len(exc_info) == 1:
    #     exc_info = exc_info[0]

    if len(exc_info) < 3:
        raise ValueError("Expected an item with >= 3 items!")

    sio = io.StringIO()
    traceback.print_exception(exc_info[0], exc_info[1], exc_info[2], None, sio)
    s = sio.getvalue()
    sio.close()
    if s[-1:] == "\n":
        s = s[:-1]
    return s



"""Configuration for pytest."""

import pytest

try:
    from PIL import Image
except ImportError:
    Image = None


@pytest.fixture(scope="session", autouse=False)
def get_icon_path(tmpdir_factory):
    """Get icon path."""
    if Image is None:
        return None
    else:
        icon = Image.new("RGB", (16, 16), color="red")
        path = str(tmpdir_factory.mktemp("data").join("img.png"))
        icon.save(path)
        return path



import numpy as np
import pytest

from qtextra.widgets.qt_button_color import QtColorButton, QtColorSwatch


@pytest.fixture
def set_qt_swatch(qtbot):
    """Setup panel"""

    def _widget(color) -> QtColorSwatch:
        widget = QtColorSwatch(initial_color=color)
        qtbot.addWidget(widget)
        return widget

    return _widget


class TestQtColorSwatch:
    @pytest.mark.parametrize("color", ("#FF0000", (255, 0, 0, 255), (1, 0, 0), (1, 0, 0, 1)))
    def test_init(self, set_qt_swatch, color):
        widget = set_qt_swatch(color)
        np.testing.assert_array_equal(widget.color, np.asarray((1.0, 0.0, 0.0, 1.0)))


@pytest.fixture
def set_qt_button(qtbot):
    """Setup panel"""

    def _widget(color) -> QtColorButton:
        widget = QtColorButton(color=color)
        qtbot.addWidget(widget)
        return widget

    return _widget


class TestQtColorButton:
    @pytest.mark.parametrize("color", ("#FF0000",))
    def test_init(self, set_qt_button, color):
        widget = set_qt_button(color)
        assert widget.color is not None



import sys

from qtpy.QtCore import Qt
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import QApplication, QComboBox, QStyle, QStyleOptionComboBox, QStylePainter


class QtElideComboBox(QComboBox):
    """A QComboBox that elides (truncates) long text with '...'."""

    def paintEvent(self, event):
        painter = QStylePainter(self)
        option = QStyleOptionComboBox()
        self.initStyleOption(option)

        # Calculate the available width for text (subtract some space for the arrow).
        text_width = self.width() - 20
        if text_width < 0:
            text_width = 0

        # Use self.font() (instead of option.font) to create the QFontMetrics.
        font_metrics = QFontMetrics(self.font())

        # Elide/truncate the combo box's current text if it's too long.
        elided_text = font_metrics.elidedText(option.currentText, Qt.TextElideMode.ElideRight, text_width)
        option.currentText = elided_text

        # Draw the combo box (frame and arrow).
        painter.drawComplexControl(QStyle.ComplexControl.CC_ComboBox, option)
        # Draw the truncated text.
        painter.drawControl(QStyle.ControlElement.CE_ComboBoxLabel, option)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    combo = QtElideComboBox()
    combo.addItems(
        [
            "Short text",
            "This is a very long text that likely needs truncation",
            "Another extremely long text example to demonstrate ellipses…",
        ]
    )
    combo.setEditable(False)
    combo.show()
    sys.exit(app.exec_())



# syntax_style for the console must be one of the supported styles from
# pygments - see here for examples https://help.farbox.com/pygments.html
import re
import typing as ty
from ast import literal_eval

from pydantic_extra_types.color import Color

from qtextra.utils.color import get_text_color, hex_to_qt_rgb, rgb_to_hex

try:
    from qtpy import QT_VERSION

    major, minor, *rest = QT_VERSION.split(".")
    use_gradients = (int(major) >= 5) and (int(minor) >= 12)
except Exception:
    use_gradients = False


increase_pattern = re.compile(r"{{\s?increase\((\w+),?\s?([-\d]+)?\)\s?}}")
decrease_pattern = re.compile(r"{{\s?decrease\((\w+),?\s?([-\d]+)?\)\s?}}")
gradient_pattern = re.compile(r"([vh])gradient\((.+)\)")
darken_pattern = re.compile(r"{{\s?darken\((\w+),?\s?([-\d]+)?\)\s?}}")
lighten_pattern = re.compile(r"{{\s?lighten\((\w+),?\s?([-\d]+)?\)\s?}}")
darken_or_lighten_pattern = re.compile(r"{{\s?darken_or_lighten\((\w+),?\s?([-\d]+)?\)\s?}}")
opacity_pattern = re.compile(r"{{\s?opacity\((\w+),?\s?([-\d]+)?\)\s?}}")
replace_pattern = re.compile(r"{{\s?replace\((\w+)\)\s?}}")


def _get_color(color: ty.Union[str, Color]) -> tuple[int, int, int]:
    if isinstance(color, str):
        if color.startswith("#"):
            color = hex_to_qt_rgb(color)
        if color.startswith("rgb("):
            color = literal_eval(color.lstrip("rgb(").rstrip(")"))
    elif isinstance(color, Color) or hasattr(color, "as_rgb_tuple"):
        color = color.as_rgb_tuple()
    return color


def decrease(font_size: str, pt: int) -> str:
    """Decrease fontsize."""
    return f"{int(font_size[:-2]) - int(pt)}pt"


def increase(font_size: str, pt: int) -> str:
    """Increase fontsize."""
    return f"{int(font_size[:-2]) + int(pt)}pt"


def color_for_background(color: ty.Union[str, Color]) -> str:
    """Return color that will stand out against background color."""
    color = _get_color(color)
    return get_text_color(rgb_to_hex(color, 1)).name()


def darken_or_lighten(color: ty.Union[str, Color], percentage=10) -> str:
    """Darken or lighten the color.

    If color is light, darken it, otherwise lighten it.
    """
    red, green, blue = _get_color(color)
    if (red * 0.299 + green * 0.587 + blue * 0.114) > 186:
        return darken(color, percentage)
    return lighten(color, percentage)


def darken(color: ty.Union[str, Color], percentage=10) -> str:
    """Darken the color."""
    red, green, blue = _get_color(color)
    ratio = 1 - float(percentage) / 100
    red = min(max(int(red * ratio), 0), 255)
    green = min(max(int(green * ratio), 0), 255)
    blue = min(max(int(blue * ratio), 0), 255)
    return f"rgb({red}, {green}, {blue})"


def lighten(color: ty.Union[str, Color], percentage=10) -> str:
    """Lighten the color."""
    red, green, blue = _get_color(color)
    ratio = float(percentage) / 100
    red = min(max(int(red + (255 - red) * ratio), 0), 255)
    green = min(max(int(green + (255 - green) * ratio), 0), 255)
    blue = min(max(int(blue + (255 - blue) * ratio), 0), 255)
    return f"rgb({red}, {green}, {blue})"


def opacity(color: ty.Union[str, Color], value=255) -> str:
    """Adjust opacity."""
    red, green, blue = _get_color(color)
    return f"rgba({red}, {green}, {blue}, {max(min(int(value), 255), 0)})"


def gradient(stops, horizontal=True) -> str:
    """Make gradient."""
    if not use_gradients:
        return stops[-1]

    if horizontal:
        grad = "qlineargradient(x1: 0, y1: 0, x2: 1, y2: 0, "
    else:
        grad = "qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1, "

    _stops = [f"stop: {n} {stop}" for n, stop in enumerate(stops)]
    grad += ", ".join(_stops) + ")"
    return grad


def template(css, **theme):
    """Generate template."""

    def _increase_match(matchobj):
        font_size, to_add = matchobj.groups()
        return increase(theme[font_size], to_add)

    def _decrease_match(matchobj):
        font_size, to_subtract = matchobj.groups()
        return decrease(theme[font_size], to_subtract)

    def _darken_match(matchobj):
        color, percentage = matchobj.groups()
        return darken(theme[color], percentage)

    def _lighten_match(matchobj):
        color, percentage = matchobj.groups()
        return lighten(theme[color], percentage)

    def _darken_or_lighten_match(matchobj):
        color, percentage = matchobj.groups()
        return darken_or_lighten(theme[color], percentage)

    def _opacity_match(matchobj):
        color, percentage = matchobj.groups()
        return opacity(theme[color], percentage)

    def _replace_match(matchobj):
        color = matchobj.groups()[0]
        return color_for_background(theme[color])

    def _gradient_match(matchobj):
        horizontal = matchobj.groups()[1] == "h"
        stops = [i.strip() for i in matchobj.groups()[1].split("-")]
        return gradient(stops, horizontal)

    for k, v in theme.items():
        if k == "name":
            k = "id"
        css = increase_pattern.sub(_increase_match, css)
        css = decrease_pattern.sub(_decrease_match, css)
        css = gradient_pattern.sub(_gradient_match, css)
        css = darken_pattern.sub(_darken_match, css)
        css = lighten_pattern.sub(_lighten_match, css)
        css = darken_or_lighten_pattern.sub(_darken_or_lighten_match, css)
        css = opacity_pattern.sub(_opacity_match, css)
        css = replace_pattern.sub(_replace_match, css)
        if isinstance(v, Color):
            v = v.as_rgb()
        css = css.replace(f"{{{{ {k} }}}}", v)
    return css



from __future__ import annotations

import operator
from decimal import Decimal

from qtpy import QtCore, QtGui
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt


class QIntOrNoneValidator(QtGui.QIntValidator):
    """Validator that accepts '' as None, and otherwise behaves as QIntValidator."""

    def validate(self, a0: str | None, a1: int) -> tuple[QtGui.QValidator.State, str, int]:
        if a0 == "":
            return QtGui.QValidator.State.Acceptable, "", a1
        return super().validate(a0, a1)


class QDoubleOrNoneValidator(QtGui.QDoubleValidator):
    """Validator that accepts '' as None, and otherwise behaves as QDoubleValidator."""

    def validate(self, a0: str | None, a1: int) -> tuple[QtGui.QValidator.State, str, int]:
        if a0 == "":
            return QtGui.QValidator.State.Acceptable, "", a1
        return super().validate(a0, a1)


class QCommaSeparatedValidator(QtGui.QValidator):
    _ChildValidator: QtGui.QValidator

    def validate(self, a0: str | None, a1: int) -> tuple[QtGui.QValidator.State, str, int]:
        if a0 == "" or a0 is None:
            return QtGui.QValidator.State.Acceptable, "", a1
        if a0.strip().endswith(","):
            if a0.strip().endswith(",,"):
                return QtGui.QValidator.State.Invalid, a0, a1
            return QtGui.QValidator.State.Intermediate, a0, a1
        state_list = [self._ChildValidator.validate(part.strip(), 0)[0] for part in a0.split(",")]
        is_valid = all(state == QtGui.QValidator.State.Acceptable for state in state_list)
        is_intermediate = all(state != QtGui.QValidator.State.Invalid for state in state_list)
        if is_valid:
            return QtGui.QValidator.State.Acceptable, a0, a1
        if is_intermediate:
            return QtGui.QValidator.State.Intermediate, a0, a1
        return QtGui.QValidator.State.Invalid, a0, a1


class QCommaSeparatedIntValidator(QCommaSeparatedValidator):
    _ChildValidator = QtGui.QIntValidator()


class QCommaSeparatedDoubleValidator(QCommaSeparatedValidator):
    _ChildValidator = QtGui.QDoubleValidator()


class QValuedLineEdit(QtW.QLineEdit):
    _validator_class: type[QIntOrNoneValidator | QDoubleOrNoneValidator]
    valueChanged = QtCore.Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setValidator(self._validator_class(self))
        self.textChanged.connect(self._on_text_changed)

    def _on_text_changed(self, text: str):
        validate_result = self.validator().validate(text, 0)
        if validate_result[0] == QtGui.QValidator.State.Acceptable:
            self.valueChanged.emit(validate_result[1])

    def sizeHint(self) -> QtCore.QSize:
        hint = super().sizeHint()
        hint.setWidth(100)  # numerical values do not need to be too wide
        return hint

    def stepUp(self, large: bool = False):
        raise NotImplementedError

    def stepDown(self, large: bool = False):
        raise NotImplementedError

    def wheelEvent(self, a0: QtGui.QWheelEvent | None) -> None:
        if a0 is not None:
            if a0.angleDelta().y() > 0:
                self.stepUp()
                a0.accept()
            elif a0.angleDelta().y() < 0:
                self.stepDown()
                a0.accept()
        return super().wheelEvent(a0)

    def keyPressEvent(self, a0: QtGui.QKeyEvent | None) -> None:
        if a0.modifiers() == Qt.KeyboardModifier.NoModifier:
            if a0.key() == Qt.Key.Key_Up:
                self.stepUp()
            elif a0.key() == Qt.Key.Key_PageUp:
                self.stepUp(large=True)
            elif a0.key() == Qt.Key.Key_Down:
                self.stepDown()
            elif a0.key() == Qt.Key.Key_PageDown:
                self.stepDown(large=True)
            else:
                return super().keyPressEvent(a0)
        else:
            return super().keyPressEvent(a0)

    def minimum(self):
        return self.validator().bottom()

    def setMinimum(self, min_):
        self.validator().setBottom(min_)

    def maximum(self):
        return self.validator().top()

    def setMaximum(self, max_):
        self.validator().setTop(max_)

    def validator(self) -> QIntOrNoneValidator | QDoubleOrNoneValidator:
        return super().validator()


class QIntLineEdit(QValuedLineEdit):
    _validator_class = QIntOrNoneValidator

    def stepUp(self, large: bool = False):
        text = self.text()
        if text == "":
            return None
        val = int(text)
        diff: int = 100 if large else 1
        self.setText(str(min(val + diff, self.validator().top())))

    def stepDown(self, large: bool = False):
        text = self.text()
        if text == "":
            return None
        val = int(text)
        diff: int = 100 if large else 1
        self.setText(str(max(val - diff, self.validator().bottom())))


class QDoubleLineEdit(QValuedLineEdit):
    """QLineEdit for double values."""

    _validator_class = QDoubleOrNoneValidator

    def stepUp(self, large: bool = False):
        return self._step_up_or_down(large, operator.add)

    def stepDown(self, large: bool = False):
        return self._step_up_or_down(large, operator.sub)

    def _step_up_or_down(self, large: bool, op):
        text = self.text()
        if text == "":
            return None
        if "e" in text:
            val_text, exp_text = text.split("e")
            if large:
                exp_dec = Decimal(exp_text)
                diff = self._calc_diff(exp_dec, False)
                exp_ = op(exp_dec, diff)
                if (
                    Decimal(val_text) * 10**exp_ > self.validator().top()
                    or Decimal(val_text) * 10**exp_ < self.validator().bottom()
                ):
                    return None
                self.setText(val_text + "e" + str(exp_))
            else:
                val_min = self.validator().bottom() / 10 ** int(exp_text)
                val_max = self.validator().top() / 10 ** int(exp_text)
                val_dec = Decimal(val_text)
                diff = self._calc_diff(val_dec, False)
                val = op(val_dec, diff)
                val = max(min(val, val_max), val_min)
                self.setText(str(val) + "e" + exp_text)
        else:
            dec = Decimal(text)
            diff = self._calc_diff(dec, large)
            val = op(dec, diff)
            val = max(min(val, self.validator().top()), self.validator().bottom())
            self.setText(str(val))

    def _calc_diff(self, dec: Decimal, large: bool):
        exponent = dec.as_tuple().exponent
        if not isinstance(exponent, int):
            return None
        ten = Decimal("10")
        diff = ten ** (exponent + 2) if large else ten**exponent
        return diff


class QCommaSeparatedIntLineEdit(QtW.QLineEdit):
    """QLineEdit for comma separated integer values."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setValidator(QCommaSeparatedIntValidator(self))


class QCommaSeparatedDoubleLineEdit(QtW.QLineEdit):
    """QLineEdit for comma separated double values."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setValidator(QCommaSeparatedDoubleValidator(self))


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe(False)
        frame.setMinimumSize(600, 600)

        ha.addWidget(QIntLineEdit(parent=frame))
        ha.addWidget(QDoubleLineEdit(parent=frame))
        ha.addWidget(QCommaSeparatedIntLineEdit(parent=frame))
        ha.addWidget(QCommaSeparatedDoubleLineEdit(parent=frame))

        frame.show()
        sys.exit(app.exec_())

    _main()



"""Overlay Message Widget - A Widget to display a temporary dismissible message over another widget."""

import sys

from qtpy.QtCore import QEvent, QPoint, QRect, QSize, Qt, Signal, Slot
from qtpy.QtGui import QPainter
from qtpy.QtWidgets import QFrame, QHBoxLayout, QLabel, QSizePolicy, QStyle, QStyleOption, QVBoxLayout, QWidget

from qtextra.helpers import make_btn
from qtextra.widgets.qt_label_icon import QtIconLabel


# noinspection PyPep8Naming
class QtOverlay(QWidget):
    """A widget positioned on top of another widget."""

    Y_OFFSET: int = 10

    def __init__(self, parent=None, alignment=Qt.AlignmentFlag.AlignCenter, **kwargs):
        super().__init__(parent, **kwargs)
        self.setContentsMargins(0, 0, 0, 0)
        self.__alignment = alignment
        self.__widget = None

    def set_widget(self, widget: QWidget):
        """Set the widget over which this overlay should be displayed (anchored)."""
        if self.__widget is not None:
            self.__widget.removeEventFilter(self)
            self.__widget.destroyed.disconnect(self.__on_destroyed)
        self.__widget = widget
        if self.__widget is not None:
            self.__widget.installEventFilter(self)
            self.__widget.destroyed.connect(self.__on_destroyed)

        if self.__widget is None:
            self.hide()
        else:
            self.__layout()

    def widget(self) -> QWidget:
        """Return the overlaid widget."""
        return self.__widget

    def setAlignment(self, alignment: Qt.AlignmentFlag):
        """Set overlay alignment."""
        if self.__alignment != alignment:
            self.__alignment = alignment
            if self.__widget is not None:
                self.__layout()

    def alignment(self) -> Qt.AlignmentFlag:
        """Return the overlay alignment."""
        return self.__alignment

    def eventFilter(self, recv, event):
        """Event filter."""
        # reimplemented
        if recv is self.__widget:
            if event.type() == QEvent.Type.Resize or event.type() == QEvent.Type.Move:
                self.__layout()
            elif event.type() == QEvent.Type.Show:
                self.show()
            elif event.type() == QEvent.Type.Hide:
                self.hide()
        return super().eventFilter(recv, event)

    def event(self, event):
        """Event."""
        # reimplemented
        if event.type() == QEvent.Type.LayoutRequest:
            self.__layout()
            return True
        else:
            return super().event(event)

    def paintEvent(self, event):
        """Paint event."""
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        self.style().drawPrimitive(QStyle.PrimitiveElement.PE_Widget, opt, painter, self)

    def showEvent(self, event):
        """Show event."""
        super().showEvent(event)
        # Force immediate re-layout on show
        self.__layout()

    def __layout(self):
        # position itself over `widget`
        # noinspection PyShadowingNames
        def _get_size(hint, minimum, maximum, policy):
            if policy == QSizePolicy.Policy.Ignored:
                return maximum
            elif policy == QSizePolicy.PolicyFlag.ExpandFlag:
                return maximum
            else:
                return max(hint, minimum)

        widget = self.__widget
        if widget is None:
            return

        alignment = self.__alignment
        policy = self.sizePolicy()

        if widget.window() is self.window() and not self.isWindow():
            if widget.isWindow():
                bounds = widget.rect()
            else:
                bounds = QRect(widget.mapTo(widget.window(), QPoint(0, 0)), widget.size())
            tl = self.parent().mapFrom(widget.window(), bounds.topLeft())
            bounds = QRect(tl, widget.size())
        else:
            if widget.isWindow():
                bounds = widget.geometry()
            else:
                bounds = QRect(widget.mapToGlobal(QPoint(0, 0)), widget.size())

            if self.isWindow():
                bounds = bounds
            else:
                bounds = QRect(self.parent().mapFromGlobal(bounds.topLeft()), bounds.size())

        sh = self.sizeHint()
        min_sh = self.minimumSizeHint()
        minsize = self.minimumSize()
        if minsize.isNull():
            minsize = min_sh
        maxsize = bounds.size().boundedTo(self.maximumSize())
        minsize = minsize.boundedTo(maxsize)
        effective_sh = sh.expandedTo(minsize).boundedTo(maxsize)

        h_policy = policy.horizontalPolicy()
        v_policy = policy.verticalPolicy()

        if not effective_sh.isValid():
            effective_sh = QSize(0, 0)
            v_policy = h_policy = QSizePolicy.Ignored

        width = _get_size(effective_sh.width(), minsize.width(), maxsize.width(), h_policy)

        height_forw = self.heightForWidth(width)
        if height_forw > 0:
            height = _get_size(height_forw, minsize.height(), maxsize.height(), v_policy)
        else:
            height = _get_size(effective_sh.height(), minsize.height(), maxsize.height(), v_policy)

        size = QSize(width, height)
        if alignment & Qt.AlignmentFlag.AlignLeft:
            x = bounds.x()
        elif alignment & Qt.AlignmentFlag.AlignRight:
            x = bounds.x() + bounds.width() - size.width()
        else:
            x = bounds.x() + max(0, bounds.width() - size.width()) // 2

        if alignment & Qt.AlignmentFlag.AlignTop:
            y = bounds.y()
        elif alignment & Qt.AlignmentFlag.AlignBottom:
            y = bounds.y() + bounds.height() - size.height()
        else:
            y = bounds.y() + max(0, bounds.height() - size.height()) // 2

        geom = QRect(QPoint(x, y + self.Y_OFFSET), size)
        self.setGeometry(geom)

    @Slot()
    def __on_destroyed(self):
        self.__widget = None
        try:
            if self.isVisible():
                self.hide()
        except RuntimeError:
            pass


class QtOverlayWidget(QFrame):
    """Widget to display message without user interaction."""

    def __init__(
        self,
        parent=None,
        text: str = "",
        **kwargs,
    ):
        super().__init__(parent, **kwargs)
        self.setFrameShape(QFrame.Shape.Box)
        self.setLineWidth(1)

        self.text_label = QLabel(text=text, wordWrap=False, textFormat=Qt.TextFormat.AutoText)
        self.text_label.setAlignment(Qt.AlignmentFlag.AlignJustify)
        if sys.platform == "darwin":
            self.text_label.setAttribute(Qt.WidgetAttribute.WA_MacSmallSize)

        row = QHBoxLayout(self)
        row.addWidget(self.text_label, stretch=True)


class QtOverlayLabel(QtOverlay):
    """Message widget that sits on top of another widget."""

    def __init__(
        self,
        parent=None,
        text="",
        alignment=Qt.AlignmentFlag.AlignTop,
        **kwargs,
    ):
        super().__init__(parent, alignment=alignment, **kwargs)
        layout = QHBoxLayout()
        layout.setContentsMargins(5, 5, 5, 5)
        self._msg_widget = QtOverlayWidget(
            parent=self,
            text=text,
        )

        layout.addWidget(self._msg_widget)
        self.setLayout(layout)


class QtMessageWidget(QFrame):
    """Widget to display simple message to the user."""

    #: Emitted when a button with an Accept role is clicked
    evt_accepted = Signal()
    #: Emitted when a button with a RejectRole is clicked
    evt_rejected = Signal()
    #: Emitted when a button with a DismissRole is clicked
    evt_dismissed = Signal()

    def __init__(
        self,
        parent=None,
        icon_name: str = "info",
        text: str = "",
        wrap_word: bool = False,
        text_format: Qt.TextFormat = Qt.TextFormat.AutoText,
        **kwargs,
    ):
        super().__init__(parent, **kwargs)
        self._text = text
        self._wordWrap = wrap_word
        self._dismissed = False

        self.setFrameShape(QFrame.Box)
        self.setLineWidth(1)

        self.icon_label = QtIconLabel(icon_name, parent=self)
        self.text_label = QLabel(text=text, wordWrap=wrap_word, textFormat=text_format)
        self.text_label.setAlignment(Qt.AlignmentFlag.AlignJustify)
        if sys.platform == "darwin":
            self.text_label.setAttribute(Qt.WidgetAttribute.WA_MacSmallSize)

        self.ok_btn = make_btn(self, "OK", tooltip="Accept")
        self.ok_btn.clicked.connect(self.on_accept)
        self.ok_btn.setVisible(False)
        self.cancel_btn = make_btn(self, "Close", tooltip="Close message")
        self.cancel_btn.clicked.connect(self.close)
        self.cancel_btn.setVisible(False)
        self.dismiss_btn = make_btn(self, "Dismiss", tooltip="Dismiss message and don't show it again in this session")
        self.dismiss_btn.clicked.connect(self.on_dismiss)
        self.dismiss_btn.setVisible(False)

        self.btn_row = QHBoxLayout()
        self.btn_row.addWidget(self.ok_btn, alignment=Qt.AlignmentFlag.AlignCenter)
        self.btn_row.addWidget(self.cancel_btn, alignment=Qt.AlignmentFlag.AlignCenter)
        self.btn_row.addWidget(self.dismiss_btn, alignment=Qt.AlignmentFlag.AlignCenter)

        row = QHBoxLayout()
        row.addWidget(self.icon_label, alignment=Qt.AlignmentFlag.AlignTop)
        row.addWidget(self.text_label, stretch=True)
        row.setSpacing(5)

        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(4, 4, 4, 4)
        self.layout.setSpacing(0)
        self.layout.addLayout(row)
        self.layout.addSpacing(5)
        self.layout.addLayout(self.btn_row)
        self.setLayout(self.layout)

    def set_buttons(
        self, ok_btn: bool = False, cancel_btn: bool = False, dismiss_btn: bool = False, ok_func=None, ok_text="OK"
    ):
        """Set buttons on the ui."""
        self.ok_btn.setVisible(ok_btn)
        self.ok_btn.setText(ok_text)
        if ok_func is not None and callable(ok_func):
            self.ok_btn.clicked.connect(ok_func)
        self.cancel_btn.setVisible(cancel_btn)
        self.dismiss_btn.setVisible(dismiss_btn)

    def display(self):
        """Display widget if it has not been dismissed before."""
        if self._dismissed:
            return
        self.setVisible(True)

    def on_accept(self):
        """Close message and accept."""
        self.hide()
        self.evt_accepted.emit()

    def on_close(self):
        """Close message and reject."""
        self.hide()
        self.evt_rejected.emit()

    def on_dismiss(self):
        """Close message and reject."""
        self.hide()
        self._dismissed = True
        self.evt_dismissed.emit()

    def dismiss(self):
        """Dismiss widget."""
        self.on_dismiss()


class QtOverlayMessage(QtOverlay):
    """Message widget that sits on top of another widget."""

    #: Emitted when a button with an Accept role is clicked
    evt_accepted = Signal()
    #: Emitted when a button with a RejectRole is clicked
    evt_rejected = Signal()
    #: Emitted when a button with a DismissRole is clicked
    evt_dismissed = Signal()

    def __init__(
        self,
        parent=None,
        text="",
        icon_name="",
        alignment=Qt.AlignmentFlag.AlignTop,
        word_wrap=False,
        can_dismiss: bool = True,
        **kwargs,
    ):
        super().__init__(parent, alignment=alignment, **kwargs)
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self._can_dismiss = can_dismiss
        self._msg_widget = QtMessageWidget(
            parent=self,
            text=text,
            icon_name=icon_name,
            wrap_word=word_wrap,
        )
        self._msg_widget.evt_accepted.connect(self.evt_accepted)
        self._msg_widget.evt_rejected.connect(self.evt_rejected)
        self._msg_widget.evt_dismissed.connect(self.evt_dismissed)

        layout.addWidget(self._msg_widget)
        self.setLayout(layout)

    @property
    def is_dismissed(self) -> bool:
        """Flag to indicate whether overlay message is dismissed."""
        return self._msg_widget._dismissed

    @property
    def is_displayed(self) -> bool:
        """Flag to indicate whether overlay message is displayed."""
        return not self._msg_widget._dismissed

    def dismiss(self):
        """Dismiss message."""
        self._msg_widget.dismiss()
        if not self._can_dismiss:
            self._msg_widget._dismissed = False

    def display(self):
        """Display message."""
        self._msg_widget.display()


class QtOverlayDismissMessage(QtOverlayMessage):
    """Message widget that sits on top of another widget."""

    def __init__(
        self,
        parent=None,
        text="",
        icon_name="",
        alignment=Qt.AlignmentFlag.AlignTop,
        word_wrap=False,
        dismiss_btn=True,
        ok_btn=False,
        ok_func=None,
        ok_text="OK",
        **kwargs,
    ):
        super().__init__(
            parent=parent, text=text, icon_name=icon_name, alignment=alignment, word_wrap=word_wrap, **kwargs
        )
        self._msg_widget.set_buttons(dismiss_btn=dismiss_btn, ok_btn=ok_btn, ok_func=ok_func, ok_text=ok_text)


if __name__ == "__main__":  # pragma: no cover
    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setLayout(ha)
    frame.setMinimumSize(400, 400)

    overlay = QtOverlayLabel(parent=frame, text="Spatial overlay text")

    # overlay = QtOverlayMessage(
    #     parent=frame,
    #     icon_name="info",
    #     text="This is a test message which should span over several lines This is a test message which
    #     should span over several lines This is a test message which should span over several lines",
    #     word_wrap=True,
    # )
    overlay.set_widget(frame)

    frame.show()
    sys.exit(app.exec_())



"""QtIcon."""

from __future__ import annotations

import typing as ty
from contextlib import suppress

from koyo.typing import PathLike
from qtpy.QtCore import QSize, Qt, Signal  # type: ignore[attr-defined]  # type: ignore[attr-defined]
from qtpy.QtGui import QEnterEvent, QPixmap, QResizeEvent
from qtpy.QtWidgets import QLabel, QToolTip, QWidget
from superqt.utils import qdebounced

from qtextra.config import THEMES
from qtextra.dialogs.qt_info_popup import InfoDialog
from qtextra.widgets._qta_mixin import QtaMixin


def make_png_label(icon_path: str, size: tuple[int, int] = (40, 40)) -> QLabel:
    """Make svg icon."""
    image = QKeepAspectLabel(None, icon_path)
    image.setMinimumSize(*size)
    return image


class QKeepAspectLabel(QLabel):
    """Keep aspect ratio label."""

    def __init__(self, parent: QWidget | None, path: PathLike):
        super().__init__(parent)
        self.path = path
        self._setPixmap()

    @qdebounced(timeout=100, leading=False)
    def _setPixmap(self) -> None:
        img = QPixmap(self.path)
        size = self.size()
        pix = img.scaled(size, Qt.AspectRatioMode.KeepAspectRatio)
        self.setPixmap(pix)

    def resizeEvent(self, event: QResizeEvent) -> None:  # type: ignore[override]
        """Resize event."""
        self._setPixmap()
        return super().resizeEvent(event)


class QtActiveIcon(QLabel):
    """Active icon that shows activity."""

    def __init__(self, which: str = "infinity", size: tuple[int, int] = (20, 20), start: bool = False):
        from qtextra.helpers import make_gif

        super().__init__()
        self.setScaledContents(True)
        self._active = False

        self.loading_movie = make_gif(which, size=size, start=start)
        if size is not None:
            self.setMaximumSize(*size)
        self.setMovie(self.loading_movie)
        self.active = start

    @property
    def active(self) -> bool:
        """Get active state."""
        return self._active

    @active.setter
    def active(self, value: bool) -> None:
        """Set active state."""
        self._active = value
        self.loading_movie.start() if value else self.loading_movie.stop()
        self.show() if value else self.hide()

    def set_active(self, active: bool) -> None:
        """Set active state."""
        self.active = active

    # Alias methods to offer Qt-like interface
    setActive = set_active


class QtIconLabel(QLabel):
    """Label with icon."""

    evt_clicked = Signal()

    def __init__(self, object_name: str, *args, **kwargs):
        super().__init__()
        self.setMouseTracking(True)
        self.setObjectName(object_name)

    def mousePressEvent(self, ev):
        """Mouse press event."""
        if ev.button() == Qt.MouseButton.LeftButton:
            self.evt_clicked.emit()
        super().mousePressEvent(ev)


class QtQtaLabel(QtIconLabel, QtaMixin):
    """Label."""

    _icon = None

    def __init__(
        self,
        *args,
        xxsmall: bool = False,
        xsmall: bool = False,
        small: bool = False,
        normal: bool = False,
        average: bool = False,
        medium: bool = False,
        large: bool = False,
        xlarge: bool = False,
        xxlarge: bool = False,
        **kwargs,
    ):
        super().__init__("", *args, **kwargs)
        self._size = QSize(28, 28)
        self.setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignHCenter)
        self.set_default_size(
            xxsmall=xxsmall,
            xsmall=xsmall,
            small=small,
            normal=normal,
            average=average,
            medium=medium,
            large=large,
            xlarge=xlarge,
            xxlarge=xxlarge,
        )
        with suppress(RuntimeError):
            THEMES.evt_theme_icon_changed.connect(self._update_qta)

    def setIcon(self, _icon) -> None:
        """Update icon."""
        self._icon = _icon
        self.setPixmap(_icon.pixmap(self._size))

    def setIconSize(self, size: QSize) -> None:
        """Set icon size."""
        self._size = size
        self.update()

    def update(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        """Update label."""
        if self._icon:
            self.setPixmap(self._icon.pixmap(self._size))
        return super().update(*args, **kwargs)


class QtQtaTooltipLabel(QtQtaLabel):
    """Label."""

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self.set_qta("help")
        self.set_average()

    def enterEvent(self, event: QEnterEvent) -> None:  # type: ignore[override]
        """Override to show tooltips instantly."""
        if self.toolTip():
            pos = self.mapToGlobal(self.contentsRect().center())
            QToolTip.showText(pos, self.toolTip(), self)
        super().enterEvent(event)


class QtQtaHelpLabel(QtQtaLabel):
    """Label."""

    _dlg: InfoDialog | None = None

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self.set_qta("help")
        self.set_average()

    def enterEvent(self, event: QEnterEvent) -> None:  # type: ignore[override]
        """Override to show tooltips instantly."""
        if self.toolTip() and not self._dlg:
            self._dlg = InfoDialog(self, self.toolTip())
            self._dlg.evt_close.connect(self._removeDialog)
            self._dlg.show_right_of_widget(self)
        super().enterEvent(event)

    def _removeDialog(self) -> None:
        """Remove dialog."""
        if self._dlg:
            self._dlg = None


class QtSeverityLabel(QtQtaLabel):
    """Severity label."""

    STATES = ("debug", "info", "success", "warning", "error", "critical")

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self._severity: str = "info"
        self.severity = "info"
        self.set_xsmall()

    @property
    def severity(self) -> str:
        """Get state."""
        return self._severity

    @severity.setter
    def severity(self, severity: str) -> None:
        self._severity = severity
        self.set_qta(severity)


class QtStateLabel(QtQtaLabel):
    """Severity label."""

    STATES = ("wait", "check", "cross", "active", "upgrade", "thread", "process", "cli")

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self._state: str = "wait"
        self.state = "wait"

    @property
    def state(self) -> str:
        """Get state."""
        return self._state

    @state.setter
    def state(self, state: str) -> None:
        self._state = state
        self.set_qta(state)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QHBoxLayout

    from qtextra.assets import QTA_MAPPING, get_icon
    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)

    lay = QHBoxLayout()
    for i, name in enumerate(QTA_MAPPING.keys()):
        qta_name, qta_kws = get_icon(name)
        qta_kws["scale_factor"] = 1
        label = QtQtaLabel()
        label.set_qta(qta_name, **qta_kws)
        label.setToolTip(f"{name} :: {qta_name}")
        label.set_large()
        lay.addWidget(label)
        if i % 20 == 0:
            ha.addLayout(lay)
            lay = QHBoxLayout()

    frame.show()
    frame.setMaximumHeight(400)
    sys.exit(app.exec_())



"""Popout."""

from __future__ import annotations

from enum import Enum
from typing import Union

from qtpy.QtCore import (
    QEasingCurve,
    QMargins,
    QObject,
    QParallelAnimationGroup,
    QPoint,
    QPropertyAnimation,
    Qt,
    Signal,
)
from qtpy.QtGui import QColor, QCursor, QIcon, QImage, QPainter, QPixmap
from qtpy.QtWidgets import QApplication, QGraphicsDropShadowEffect, QHBoxLayout, QVBoxLayout, QWidget

import qtextra.helpers as hp
from qtextra.config import is_dark
from qtextra.utils.wrap import TextWrap
from qtextra.widgets.qt_label_image import QImageLabel


class PopoutAnimationType(Enum):
    """Popout animation type."""

    SLIDE_UP = 0
    SLIDE_DOWN = 1
    SLIDE_LEFT = 2
    SLIDE_RIGHT = 3
    FADE_IN = 4
    NONE = 5


class PopoutViewBase(QWidget):
    """Popout view base class."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)

    def addSpacing(self, size):
        raise NotImplementedError("Must implement method")

    def addWidget(self, widget: QWidget, stretch=0, align=Qt.AlignmentFlag.AlignLeft):
        raise NotImplementedError

    def background_color(self):
        """Return the background color."""
        return QColor(40, 40, 40) if is_dark() else QColor(248, 248, 248)

    def border_color(self):
        """Return the border color."""
        return QColor(0, 0, 0, 45) if is_dark() else QColor(0, 0, 0, 17)

    def paintEvent(self, e):
        """Paint event."""
        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing)

        painter.setBrush(self.background_color())
        painter.setPen(self.border_color())

        rect = self.rect().adjusted(1, 1, -1, -1)
        painter.drawRoundedRect(rect, 8, 8)


class PopoutView(PopoutViewBase):
    """Popout view.

    Parameters
    ----------
    title: str
        the title of teaching tip

    content: str
        the content of teaching tip

    icon: InfoBarIcon | QIcon | str
        the icon of teaching tip

    image: str | QPixmap | QImage
        the image of teaching tip

    is_closable: bool
        whether to show the close button

    parent: QWidget
        parent widget
    """

    evt_closed = Signal()

    def __init__(
        self,
        title: str,
        content: str,
        icon: Union[QIcon, str] = None,
        image: Union[str, QPixmap, QImage] = None,
        is_closable=False,
        parent=None,
    ):
        super().__init__(parent=parent)

        self.icon = icon
        self.title = title
        self.image = image
        self.content = content
        self.is_closable = is_closable

        self.titleLabel = hp.make_label(self, title, bold=True, object_name="titleLabel")
        self.contentLabel = hp.make_label(self, title, object_name="contentLabel")

        # add icon widget
        self.iconWidget = hp.make_qta_label(self, icon or "")
        if not self.title or not self.content:
            self.iconWidget.setFixedHeight(36)

        self.imageLabel = QImageLabel(self)
        self.imageLabel.setImage(self.image)

        self.closeButton = hp.make_qta_btn(self, "cross", func=self.evt_closed, average=True, hide=self.is_closable)
        self.closeButton.clicked.connect(self.evt_closed)

        self.titleLabel.setVisible(bool(self.title))
        self.contentLabel.setVisible(bool(self.content))
        self.iconWidget.setHidden(self.icon is None)

        self.viewLayout = QHBoxLayout()
        self.viewLayout.setSpacing(4)

        self.widgetLayout = QVBoxLayout()
        self.widgetLayout.setContentsMargins(0, 8, 0, 8)
        self.widgetLayout.setSpacing(0)

        self.vBoxLayout = QVBoxLayout(self)
        self.vBoxLayout.setContentsMargins(1, 1, 1, 1)
        self.vBoxLayout.setSpacing(0)
        self.vBoxLayout.addLayout(self.viewLayout)

        self.viewLayout.addWidget(self.iconWidget, 0, Qt.AlignmentFlag.AlignVCenter)

        # add text
        self._adjustText()
        self.widgetLayout.addWidget(self.titleLabel)
        self.widgetLayout.addWidget(self.contentLabel)
        self.viewLayout.addLayout(self.widgetLayout)

        # add close button
        self.closeButton.setVisible(self.is_closable)
        self.viewLayout.addWidget(self.closeButton, 0, Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)

        # adjust content margins
        margins = QMargins(6, 5, 6, 5)
        margins.setLeft(20 if not self.icon else 5)
        margins.setRight(20 if not self.is_closable else 6)
        self.viewLayout.setContentsMargins(margins)

        # add image
        self._adjustImage()
        self._addImageToLayout()

    def addSpacing(self, size):
        """Add spacing to view."""
        self.widgetLayout.addSpacing(size)

    def addWidget(self, widget: QWidget, stretch=0, align=Qt.AlignmentFlag.AlignLeft):
        """Add widget to view."""
        self.widgetLayout.addSpacing(8)
        self.widgetLayout.addWidget(widget, stretch, align)

    def _addImageToLayout(self):
        self.imageLabel.setBorderRadius(8, 8, 0, 0)
        self.imageLabel.setHidden(self.imageLabel.isNull())
        self.vBoxLayout.insertWidget(0, self.imageLabel)

    def _adjustText(self):
        w = min(900, QApplication.screenAt(QCursor.pos()).geometry().width() - 200)

        # adjust title
        chars = max(min(w / 10, 120), 30)
        self.titleLabel.setText(TextWrap.wrap(self.title, chars, False)[0])

        # adjust content
        chars = max(min(w / 9, 120), 30)
        self.contentLabel.setText(TextWrap.wrap(self.content, chars, False)[0])

    def _adjustImage(self):
        w = self.vBoxLayout.sizeHint().width() - 2
        self.imageLabel.scaledToWidth(w)

    def showEvent(self, e):
        """Show event."""
        super().showEvent(e)
        self._adjustImage()
        self.adjustSize()


class QtPopout(QWidget):
    """Popout."""

    evt_closed = Signal()

    def __init__(self, view: PopoutViewBase, parent=None, close_on_delete=True):
        super().__init__(parent=parent)
        self.view = view
        self.hBoxLayout = QHBoxLayout(self)
        self.aniManager = None  # type: PopoutAnimationManager
        self.close_on_delete = close_on_delete

        self.hBoxLayout.setContentsMargins(15, 8, 15, 20)
        self.hBoxLayout.addWidget(self.view)
        self.setShadowEffect()

        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setWindowFlags(
            Qt.WindowType.Popup | Qt.WindowType.FramelessWindowHint | Qt.WindowType.NoDropShadowWindowHint
        )

    def setShadowEffect(self, blurRadius=35, offset=(0, 8)):
        """Add shadow to dialog."""
        color = QColor(0, 0, 0, 80 if is_dark() else 30)
        self.shadowEffect = QGraphicsDropShadowEffect(self.view)
        self.shadowEffect.setBlurRadius(blurRadius)
        self.shadowEffect.setOffset(*offset)
        self.shadowEffect.setColor(color)
        self.view.setGraphicsEffect(None)
        self.view.setGraphicsEffect(self.shadowEffect)

    def closeEvent(self, e):
        """Close event."""
        if self.close_on_delete:
            self.deleteLater()

        super().closeEvent(e)
        self.evt_closed.emit()

    def showEvent(self, e):
        # fixes #780
        self.activateWindow()
        super().showEvent(e)

    def exec(self, pos: QPoint, animation_type=PopoutAnimationType.SLIDE_UP):
        """Show calendar view."""
        self.aniManager = PopoutAnimationManager.make(animation_type, self)
        self.show()
        self.aniManager.exec(pos)

    @classmethod
    def _init(
        cls,
        view: PopoutViewBase,
        target: Union[QWidget, QPoint] = None,
        parent=None,
        animation_type: PopoutAnimationType = PopoutAnimationType.SLIDE_UP,
        close_on_delete: bool = True,
    ):
        """Create and show a Popout.

        Parameters
        ----------
        view: PopoutViewBase
            Popout view

        target: QWidget | QPoint
            the target widget or position to show Popout

        parent: QWidget
            parent window

        animation_type: PopoutAnimationType
            Popout animation type

        close_on_delete: bool
            whether delete Popout automatically when Popout is evt_closed
        """
        w = cls(view, parent, close_on_delete)

        if target is None:
            return w

        # show Popout first so that we can get the correct size
        w.show()

        # move Popout to the top of target
        if isinstance(target, QWidget):
            target = PopoutAnimationManager.make(animation_type, w).position(target)

        w.exec(target, animation_type)
        return w

    @classmethod
    def init(
        cls,
        title: str,
        content: str,
        icon: Union[QIcon, str] = None,
        image: Union[str, QPixmap, QImage] = None,
        is_closable=False,
        target: Union[QWidget, QPoint] = None,
        parent=None,
        animation_type: PopoutAnimationType = PopoutAnimationType.SLIDE_UP,
        close_on_delete=True,
    ):
        """Create and show a Popout using the default view.

        Parameters
        ----------
        title: str
            the title of teaching tip

        content: str
            the content of teaching tip

        icon: InfoBarIcon | QIcon | str
            the icon of teaching tip

        image: str | QPixmap | QImage
            the image of teaching tip

        is_closable: bool
            whether to show the close button

        target: QWidget | QPoint
            the target widget or position to show Popout

        parent: QWidget
            parent window

        animation_type: PopoutAnimationType
            Popout animation type

        close_on_delete: bool
            whether delete Popout automatically when Popout is evt_closed
        """
        view = PopoutView(title, content, icon, image, is_closable)
        w = cls._init(view, target, parent, animation_type, close_on_delete)
        view.evt_closed.connect(w.close)
        return w

    def fade_out(self) -> None:
        self.fadeOutAni = QPropertyAnimation(self, b"windowOpacity", self)
        self.fadeOutAni.finished.connect(self.close)
        self.fadeOutAni.setStartValue(1)
        self.fadeOutAni.setEndValue(0)
        self.fadeOutAni.setDuration(120)
        self.fadeOutAni.start()


class PopoutAnimationManager(QObject):
    """Popout animation manager."""

    managers = {}

    def __init__(self, popout: QtPopout):
        super().__init__()
        self._popout = popout
        self.aniGroup = QParallelAnimationGroup(self)
        self.slideAni = QPropertyAnimation(popout, b"pos", self)
        self.opacityAni = QPropertyAnimation(popout, b"windowOpacity", self)

        self.slideAni.setDuration(187)
        self.opacityAni.setDuration(187)

        self.opacityAni.setStartValue(0)
        self.opacityAni.setEndValue(1)

        self.slideAni.setEasingCurve(QEasingCurve.Type.OutQuad)
        self.opacityAni.setEasingCurve(QEasingCurve.Type.OutQuad)
        self.aniGroup.addAnimation(self.slideAni)
        self.aniGroup.addAnimation(self.opacityAni)

    @classmethod
    def register(cls, name):
        """Register menu animation manager.

        Parameters
        ----------
        name: Any
            the name of manager, it should be unique
        """

        def wrapper(Manager):
            if name not in cls.managers:
                cls.managers[name] = Manager

            return Manager

        return wrapper

    def exec(self, pos: QPoint):
        """Start animation."""
        raise NotImplementedError

    def _adjustPosition(self, pos):
        rect = hp.get_current_screen_geometry()
        w, h = self._popout.sizeHint().width() + 5, self._popout.sizeHint().height()
        x = max(rect.left(), min(pos.x(), rect.right() - w))
        y = max(rect.top(), min(pos.y() - 4, rect.bottom() - h + 5))
        return QPoint(x, y)

    def position(self, target: QWidget):
        """Return the top left position relative to the target."""
        raise NotImplementedError

    @classmethod
    def make(cls, animation_type: PopoutAnimationType, popup: QtPopout) -> PopoutAnimationManager:
        """Mask animation manager."""
        if animation_type not in cls.managers:
            raise ValueError(f"`{animation_type}` is an invalid animation type.")

        return cls.managers[animation_type](popup)


@PopoutAnimationManager.register(PopoutAnimationType.SLIDE_UP)
class PullUpPopoutAnimationManager(PopoutAnimationManager):
    """Pull up Popout animation manager."""

    def position(self, target: QWidget):
        w = self._popout
        pos = target.mapToGlobal(QPoint())
        x = pos.x() + target.width() // 2 - w.sizeHint().width() // 2
        y = pos.y() - w.sizeHint().height() + w.layout().contentsMargins().bottom()
        return QPoint(x, y)

    def exec(self, pos: QPoint):
        pos = self._adjustPosition(pos)
        self.slideAni.setStartValue(pos + QPoint(0, 8))
        self.slideAni.setEndValue(pos)
        self.aniGroup.start()


@PopoutAnimationManager.register(PopoutAnimationType.SLIDE_DOWN)
class DropDownPopoutAnimationManager(PopoutAnimationManager):
    """Drop down Popout animation manager."""

    def position(self, target: QWidget):
        w = self._popout
        pos = target.mapToGlobal(QPoint(0, target.height()))
        x = pos.x() + target.width() // 2 - w.sizeHint().width() // 2
        y = pos.y() - w.layout().contentsMargins().top() + 8
        return QPoint(x, y)

    def exec(self, pos: QPoint):
        pos = self._adjustPosition(pos)
        self.slideAni.setStartValue(pos - QPoint(0, 8))
        self.slideAni.setEndValue(pos)
        self.aniGroup.start()


@PopoutAnimationManager.register(PopoutAnimationType.SLIDE_LEFT)
class SlideLeftPopoutAnimationManager(PopoutAnimationManager):
    """Slide left Popout animation manager."""

    def position(self, target: QWidget):
        w = self._popout
        pos = target.mapToGlobal(QPoint(0, 0))
        x = pos.x() - w.sizeHint().width() + 8
        y = pos.y() - w.sizeHint().height() // 2 + target.height() // 2 + w.layout().contentsMargins().top()
        return QPoint(x, y)

    def exec(self, pos: QPoint):
        pos = self._adjustPosition(pos)
        self.slideAni.setStartValue(pos + QPoint(8, 0))
        self.slideAni.setEndValue(pos)
        self.aniGroup.start()


@PopoutAnimationManager.register(PopoutAnimationType.SLIDE_RIGHT)
class SlideRightPopoutAnimationManager(PopoutAnimationManager):
    """Slide right Popout animation manager."""

    def position(self, target: QWidget):
        w = self._popout
        pos = target.mapToGlobal(QPoint(0, 0))
        x = pos.x() + target.width() - 8
        y = pos.y() - w.sizeHint().height() // 2 + target.height() // 2 + w.layout().contentsMargins().top()
        return QPoint(x, y)

    def exec(self, pos: QPoint):
        pos = self._adjustPosition(pos)
        self.slideAni.setStartValue(pos - QPoint(8, 0))
        self.slideAni.setEndValue(pos)
        self.aniGroup.start()


@PopoutAnimationManager.register(PopoutAnimationType.FADE_IN)
class FadeInPopoutAnimationManager(PopoutAnimationManager):
    """Fade in Popout animation manager."""

    def position(self, target: QWidget):
        w = self._popout
        pos = target.mapToGlobal(QPoint())
        x = pos.x() + target.width() // 2 - w.sizeHint().width() // 2
        y = pos.y() - w.sizeHint().height() + w.layout().contentsMargins().bottom()
        return QPoint(x, y)

    def exec(self, pos: QPoint):
        self._popout.move(self._adjustPosition(pos))
        self.aniGroup.removeAnimation(self.slideAni)
        self.aniGroup.start()


@PopoutAnimationManager.register(PopoutAnimationType.NONE)
class DummyPopoutAnimationManager(PopoutAnimationManager):
    """Dummy Popout animation manager."""

    def exec(self, pos: QPoint):
        """Start animation."""
        self._popout.move(self._adjustPosition(pos))

    def position(self, target: QWidget):
        """Return the top left position relative to the target."""
        m = self._popout.hBoxLayout.contentsMargins()
        return target.mapToGlobal(QPoint(-m.left(), -self._popout.sizeHint().height() + m.bottom() - 8))


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys
        from random import choice

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe()

        def _popup():
            QtPopout.init(
                "Hello World",
                "Here is some text that should be displayed below the title",
                parent=frame,
                animation_type=choice(list(PopoutAnimationType)),
                target=btn,
                is_closable=True,
            )

        def _popup2():
            class MyPopoutView(PopoutView):
                def __init__(self, title, content, icon=None, image=None, parent=None):
                    super().__init__(title, content, icon, image, True, parent=parent)
                    self.addWidget(hp.make_btn(self, "Button 1"))
                    self.addWidget(hp.make_btn(self, "Button 2"))
                    self.addWidget(hp.make_btn(self, "Button 3"))

            pop = MyPopoutView(
                "Hello World",
                "Here is some text that should be displayed below the title",
                parent=frame,
            )
            pop.show()

        btn = hp.make_btn(frame, "Show Popout", func=_popup)
        ha.addWidget(btn)
        btn2 = hp.make_btn(frame, "Show Popout with buttons", func=_popup2)
        ha.addWidget(btn2)

        frame.show()
        sys.exit(app.exec_())

    _main()  # type: ignore[no-untyped-call]



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QCheckBox


class QtTriCheckBox(QCheckBox):
    """Custom checkbox that can be in three states but only toggles true/false when user clicks."""

    def __init__(self, *args, **kwargs):
        QCheckBox.__init__(self, *args, **kwargs)
        self.clicked.connect(self.on_clicked)

    def on_clicked(self):
        """On clicked."""
        state = self.checkState()
        new_state = Qt.CheckState.Unchecked if state == Qt.CheckState.Checked else Qt.CheckState.Checked
        self.setCheckState(new_state)



"""Simple indicator widget."""

from __future__ import annotations

from qtpy.QtCore import QEasingCurve, QPoint, QPropertyAnimation, Slot
from qtpy.QtGui import QPainter
from qtpy.QtWidgets import QGraphicsOpacityEffect, QSizePolicy, QWidget

INDICATOR_TYPES = {"success": "success", "warning": "warning", "active": "progress"}
DEFAULT_START_OPACITY = 1.0
DEFAULT_END_OPACITY = 0.2
DEFAULT_PULSE_RATE = 1000
DEFAULT_N_LOOPS = 5


class QtIndicator(QWidget):
    """Small indicator widget that flashes occasionally."""

    # states: success, warning, active, none
    START_OPACITY = DEFAULT_START_OPACITY
    END_OPACITY = DEFAULT_END_OPACITY
    PULSE_RATE = DEFAULT_PULSE_RATE
    N_LOOPS = DEFAULT_N_LOOPS

    def __init__(self, parent=None, max_size=None):
        super().__init__(parent=parent)
        self.setProperty("state", "none")
        self.setProperty("active", "False")
        if max_size:
            self.setMaximumSize(*max_size)
            self.setMinimumSize(*max_size)

        self.opacity = QGraphicsOpacityEffect()
        self.setGraphicsEffect(self.opacity)
        self.opacity_anim = QPropertyAnimation(self.opacity, b"opacity", self)
        self.opacity_anim.currentLoopChanged.connect(self._loop_update)
        self.opacity_anim.finished.connect(self.stop_pulse)

        self.setContentsMargins(2, 2, 2, 2)
        self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)

    @property
    def state(self):
        """Get state."""
        return self.property("state")

    @state.setter
    def state(self, value: str):
        self.setProperty("state", value)

    @property
    def active(self):
        """Get active state."""
        return self.property("active")

    @active.setter
    def active(self, value: bool):
        self.setProperty("active", str(value))
        self.start_pulse() if value else self.stop_pulse()

    @Slot(int)
    def _loop_update(self, loop: int | None = None):
        """Reverse pulse direction for nicer visual effect."""
        if loop is None:
            loop = self.opacity_anim.currentLoop()
        start, end = (self.START_OPACITY, self.END_OPACITY) if loop % 2 == 0 else (self.END_OPACITY, self.START_OPACITY)
        self.opacity_anim.setStartValue(start)
        self.opacity_anim.setEndValue(end)

    def paintEvent(self, *args):
        """Paint event."""
        super().paintEvent(*args)
        # default paint
        width = self.rect().width()
        height = self.rect().height()
        pos = QPoint(width - int(width / 2) - 5, height - int(height / 4) - 5)

        paint = QPainter(self)
        pen = paint.pen()
        paint.setBrush(pen.brush())
        paint.drawEllipse(pos, int(width / 4), int(height / 4))

    def start_pulse(self):
        """Start pulsating."""
        self.opacity_anim.setEasingCurve(QEasingCurve.Type.Linear)
        self.opacity_anim.setDuration(self.PULSE_RATE)
        self.opacity_anim.setStartValue(self.START_OPACITY)
        self.opacity_anim.setEndValue(self.END_OPACITY)
        self.opacity_anim.setLoopCount(self.N_LOOPS)
        self.opacity_anim.start()

    def stop_pulse(self):
        """Stop pulsating."""
        self.opacity_anim.stop()
        self.opacity.setOpacity(1.0)

    def temporary_pulse(self, duration: int = 1000, pulse: int = DEFAULT_PULSE_RATE):
        """Pulse for a short duration."""
        if self.opacity_anim.state() == QPropertyAnimation.State.Running:
            self.opacity_anim.stop()
        self.opacity_anim.setEasingCurve(QEasingCurve.Type.Linear)
        self.opacity_anim.setDuration(pulse)
        self.opacity_anim.setStartValue(self.START_OPACITY)
        self.opacity_anim.setEndValue(self.END_OPACITY)
        self.opacity_anim.setLoopCount(max(1, round(duration / pulse)))
        self.opacity_anim.finished.connect(self.hide)
        self.opacity_anim.start()
        self.show()


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe(False)
        frame.setMinimumSize(600, 600)

        btn2 = QtIndicator(parent=frame)
        btn2.setMaximumSize(16, 16)
        btn2.state = "warning"
        btn2.start_pulse()
        ha.addWidget(btn2)

        btn2 = QtIndicator(parent=frame)
        btn2.setMaximumSize(20, 20)
        btn2.state = "success"
        btn2.start_pulse()
        ha.addWidget(btn2)

        btn2 = QtIndicator(parent=frame)
        btn2.setMaximumSize(80, 80)
        btn2.state = "active"
        btn2.start_pulse()
        ha.addWidget(btn2)

        frame.show()
        sys.exit(app.exec_())

    _main()



"""Events emitter."""

from qtpy.QtCore import QObject, Signal

try:
    from qtextra.utils.notifications import Notification
except ImportError:
    Notification = object


class Events(QObject):
    """Events emitter."""

    # emitted when help view is requested
    evt_help_request = Signal(str)

    # logger/information events
    evt_msg_info = Signal(str)
    evt_msg_success = Signal(str)
    evt_msg_warning = Signal(str)
    evt_msg_error = Signal(str)
    evt_msg_critical = Signal(str)

    # exception
    evt_exception = Signal(tuple)

    # notification events
    evt_notification_action = Signal(str, tuple)
    evt_notification_dismiss = Signal()
    evt_notification = Signal(Notification)

    # splash screen events
    evt_splash_msg = Signal(str)  # message to be displayed
    evt_splash_close = Signal()

    # Statusbar events
    evt_statusbar_help = Signal(object)
    evt_statusbar_status = Signal(object)

    # exit events
    evt_force_exit = Signal()

    def emit_evt_statusbar_status(self, obj):
        """Emit statusbar status event."""
        self.evt_statusbar_status.emit(obj)

    def emit_evt_statusbar_help(self, obj):
        """Emit statusbar help event."""
        self.evt_statusbar_help.emit(obj)


EVENTS: Events = Events()



import sys
import typing as ty
from functools import lru_cache
from pathlib import Path

from koyo.utilities import running_as_pyinstaller_app
from loguru import logger


def running_under_pytest() -> bool:
    """Return True if currently running under py.test.

    This function is used to do some adjustment for testing. The environment
    variable ORIGAMI_PYTEST is defined in conftest.py.
    """
    import os

    return bool(os.environ.get("QTEXTRA_PYTEST"))


def get_docs_path() -> Path:
    """Get path to docs directory."""
    base_path = Path(sys.executable).parent

    if running_as_pyinstaller_app():
        docs_path = base_path
    else:
        docs_path = base_path.parent
    return docs_path


def check_url(url: str) -> bool:
    """Parse typical URL.

    See: https://stackoverflow.com/a/50352868
    """
    from urllib.parse import urljoin, urlparse

    final_url = urlparse(urljoin(url, "/"))
    return all([final_url.scheme, final_url.netloc, final_url.path]) and len(final_url.netloc.split(".")) > 1


def memory_usage() -> float:
    """
    Return physical memory usage (float)
    Requires the cross-platform psutil (>=v0.3) library
    (https://github.com/giampaolo/psutil).
    """
    import psutil

    # This is needed to avoid a deprecation warning error with
    # newer psutil versions
    try:
        percent = psutil.virtual_memory().percent
    except Exception:
        percent = psutil.phymem_usage().percent
    return percent


def process_memory_usage() -> float:
    """Return process memory usage (float)."""
    import psutil

    process = psutil.Process()
    return process.memory_percent()


@lru_cache(maxsize=2)
def get_system_info(as_html=False) -> str:
    """Gathers relevant module versions for troubleshooting purposes.

    Parameters
    ----------
    as_html : bool
        if True, info will be returned as HTML, suitable for a QTextEdit widget
    """
    import platform
    import sys

    from qtextra import __version__ as qtextra_version

    sys_version = sys.version.replace("\n", " ")

    text = f"<b>Python</b>: {sys_version}<br>"
    text += f"<b>Platform</b>: {platform.platform()}<br><br>"
    text += f"<b>qtextra</b>: {qtextra_version}<br>"

    try:
        from qtpy import API_NAME, PYQT_VERSION, PYSIDE_VERSION, QtCore

        if API_NAME == "PySide2":
            API_VERSION = PYSIDE_VERSION
        elif API_NAME == "PyQt5":
            API_VERSION = PYQT_VERSION
        else:
            API_VERSION = ""

        text += f"<b>Qt</b>: {QtCore.__version__}<br>"
        text += f"<b>{API_NAME}</b>: {API_VERSION}<br>"

    except Exception as e:
        text += f"<b>Qt</b>: Import failed ({e})<br>"

    modules = (
        ("numpy", "NumPy"),
        ("scipy", "SciPy"),
        ("vispy", "VisPy"),
        ("napari", "Napari"),
        ("sklearn", "Scikit-Learn"),
        ("xgboost", "XGBoost"),
    )
    loaded = {}
    for module, name in modules:
        try:
            loaded[module] = __import__(module)
            text += f"<b>{name}</b>: {loaded[module].__version__}<br>"
        except Exception as e:
            text += f"<b>{name}</b>: Import failed ({e})<br>"

    text += "<br><b>OpenGL:</b><br>"

    if loaded.get("vispy", False):
        sys_info_text = (
            "<br>".join([loaded["vispy"].sys_info().split("\n")[index] for index in [-4, -3]])
            .replace("'", "")
            .replace("<br>", "<br>  - ")
        )
        text += f"  - {sys_info_text}<br>"
    else:
        text += "  - failed to load vispy"

    text += "<br><b>Screens:</b><br>"

    try:
        from qtpy.QtGui import QGuiApplication

        screen_list = QGuiApplication.screens()
        for i, screen in enumerate(screen_list, start=1):
            text += (
                f"  - screen {i}: resolution {screen.geometry().width()}x{screen.geometry().height()},"
                f" scale {screen.devicePixelRatio()}<br>"
            )
    except Exception as e:
        text += f"  - failed to load screen information {e}"

    if not as_html:
        text = text.replace("<br>", "\n").replace("<b>", "").replace("</b>", "")
    return text


class Connectable(ty.Protocol):
    """Protocol for connectable objects."""

    def connect(self, func: ty.Callable) -> ty.Any:
        """Connect function."""

    def disconnect(self, func: ty.Callable) -> ty.Any:
        """Disconnect function."""


def connect(
    connectable: Connectable, func: ty.Callable, state: bool = True, source: str = "", silent: bool = False
) -> None:
    """Function that connects/disconnects."""
    try:
        connectable_func = connectable.connect if state else connectable.disconnect
        connectable_func(func)
    except Exception as exc:
        if not silent:
            text = (
                f"Failed to {'' if state else 'dis'}connect function; error='{exc}'; func={func};"
                f" connectable={connectable}"
            )
            if source:
                text += f"; source={source}"
            logger.debug(text)



from __future__ import annotations

import typing as ty

from qtpy.QtCore import QEvent, Signal
from qtpy.QtGui import QPainter
from qtpy.QtWidgets import QStyle, QStyleOption, QWidget

from qtextra import helpers as hp


class QtDragWidget(QWidget):
    """Drag widget."""

    evt_dropped = Signal("QEvent")

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self.setAutoFillBackground(True)
        self.setAcceptDrops(True)

    def _update_property(self, prop: str, value: bool) -> None:
        """Update properties of widget to update style."""
        self.setProperty(prop, value)
        hp.polish_widget(self)

    def dragEnterEvent(self, event: QEvent) -> None:
        """Override Qt method.

        Provide style updates on event.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self._update_property("drag", True)
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        """Override Qt method.

        Provide style updates on event.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self._update_property("drag", False)

    def dropEvent(self, event):
        """Override Qt method.

        Provide style updates on event and emit the drop event.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self._update_property("drag", False)
        self.evt_dropped.emit(event)

    def paintEvent(self, event):
        """Override Qt method.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        option = QStyleOption()
        option.initFrom(self)
        p = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, option, p, self)



"""HTML viewer."""

import os
import typing as ty

from qtpy.QtCore import QStandardPaths
from qtpy.QtWidgets import QWidget

try:
    from qtpy.QtWebEngineWidgets import QWebEngineView  # type: ignore[attr-defined]

    QWebView = None
except ImportError:
    QWebEngineView = None
    try:
        from qtpy.QtNetwork import QNetworkDiskCache
        from qtpy.QtWebKitWidgets import QWebView  # type: ignore[no-redef]

    except ImportError:
        QWebView = None


def make_html_viewer(parent: QWidget) -> ty.Any:
    """Make viewier."""
    if QWebEngineView is not None:
        panel_help = QWebEngineView(parent)
    elif QWebView is not None:
        panel_help = QWebView(parent)
        manager = panel_help.page().networkAccessManager()
        cache = QNetworkDiskCache()
        cache_dir = os.path.join(
            QStandardPaths.writableLocation(QStandardPaths.StandardLocation.CacheLocation), "help", "help-view-cache"
        )
        cache.setCacheDirectory(cache_dir)
        manager.setCache(cache)
    else:
        panel_help = None
    return panel_help



"""Various mixin classes that can be integrated into other widgets."""

from __future__ import annotations

import time
import typing as ty
from contextlib import contextmanager
from functools import partial
from pathlib import Path

from koyo.timer import report_time
from loguru import logger
from qtpy.QtCore import QTimer, Signal  # type: ignore[attr-defined]
from qtpy.QtGui import QCloseEvent
from qtpy.QtWidgets import QHBoxLayout, QLabel, QPushButton, QWidget

import qtextra.helpers as hp
from qtextra.config import EVENTS, get_settings
from qtextra.utils.utilities import check_url, get_docs_path

# Documentation directory
DOC_DIR = Path(get_docs_path())


class DocumentationMixin:
    """Documentation mixin."""

    DOC_HTML_LINK: str = ""

    parent: ty.Callable[..., QWidget]

    def _make_info_layout(
        self, align_right: bool = True, html_link: str = "", parent: QWidget | None = None
    ) -> ty.Tuple[QPushButton, QHBoxLayout]:
        """Make info button."""
        if not html_link:
            html_link = self.DOC_HTML_LINK

        info_btn = hp.make_qta_btn(
            parent if parent is not None else self.parent(),
            "help",
            tooltip="Click here to see more information about this panel...",
        )
        info_btn.clicked.connect(partial(self._open_info_link, html_link))

        layout = QHBoxLayout()
        layout.addWidget(info_btn)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        if align_right:
            layout.insertSpacerItem(0, hp.make_h_spacer())
        else:
            layout.addSpacerItem(hp.make_h_spacer())
        return info_btn, layout

    @staticmethod
    def _open_info_link(html_link: str) -> None:
        """Open link."""
        # local docs
        if html_link.startswith("docs/"):
            html_link = (DOC_DIR / html_link).as_uri()

        if html_link.startswith("file:///") or check_url(html_link):
            EVENTS.evt_help_request.emit(html_link)
        else:
            EVENTS.evt_msg_warning.emit("The provided link is not valid")


class ConfigMixin:
    """Configuration mixin."""

    _initialized_config: bool = False
    _is_setting_config: bool = False

    @contextmanager
    def setting_config(self) -> ty.Generator[None, None, None]:
        """Disable updates by temporarily setting the `_is_setting_config` flag."""
        self._is_setting_config = True
        yield
        self._is_setting_config = False

    def on_set_from_config(self) -> None:
        """Init from config."""
        with self.setting_config():
            self._on_set_from_config(get_settings())  # type: ignore[no-untyped-call]
        self._initialized_config = True

    def _on_set_from_config(self, settings: ty.Any | None = None) -> None:
        """Bind events."""


class TimerMixin:
    """Timer mixin."""

    def _add_periodic_timer(self, interval: int, fcn: ty.Callable | None, start: bool = True) -> QTimer:
        """Create timer to execute some action."""
        timer = QTimer(self)  # type: ignore[arg-type]
        timer.setInterval(interval)
        if fcn:
            timer.timeout.connect(fcn)

        if start:
            timer.start()
        logger.debug(f"Added periodic timer event that runs every {interval / 1000}s")
        return timer

    def _add_single_shot_timer(self, delay: int, fcn: ty.Callable) -> QTimer:
        timer = QTimer(self)  # type: ignore[arg-type]
        timer.singleShot(delay, fcn)
        return timer

    @contextmanager  # type: ignore[arg-type]
    def measure_time(
        self, message: str = "Task took", func: ty.Callable = logger.trace
    ) -> ty.Generator[None, None, None]:
        """Measure time."""
        t_start = time.time()
        yield
        func(f"{message} {report_time(t_start)}")


class MinimizeMixin:
    """Mixin class to enable hiding of popup."""

    _make_move_handle: ty.Callable[..., ty.Any]
    hide: ty.Callable[..., ty.Any]
    clearFocus: ty.Callable[..., ty.Any]

    def _make_hide_handle(self) -> tuple[QPushButton, QHBoxLayout]:
        hide_btn = hp.make_qta_btn(
            self,  # type: ignore[arg-type]
            "minimise",
            tooltip="Click here to minimize the popup window",
        )
        hide_btn.clicked.connect(self.on_hide)

        hide_layout = self._make_move_handle()
        hide_layout.addWidget(hide_btn)
        return hide_btn, hide_layout

    def on_hide(self) -> None:
        """Hide."""
        self.hide()
        self.clearFocus()

    def closeEvent(self, event: QCloseEvent) -> None:
        """Hide."""
        self.on_hide()
        event.ignore()


# noinspection PyUnresolvedReferences
class CloseMixin:
    """Mixin class to enable closing of popup."""

    HIDE_WHEN_CLOSE: bool = False
    _title_label: QLabel

    close: ty.Callable[..., ty.Any]
    _make_move_handle: ty.Callable[..., ty.Any]

    def _make_close_handle(self, title: str = "") -> tuple[QPushButton, QHBoxLayout]:
        close_btn = hp.make_qta_btn(
            self,
            "cross",
            tooltip="Click here to close the popup window",
            normal=True,
        )
        close_btn.clicked.connect(self.close)

        close_layout = self._make_move_handle()
        close_layout.addWidget(close_btn)
        self._title_label.setText(title)
        return close_btn, close_layout

    def _make_hide_handle(self, title: str = "") -> tuple[QPushButton, QHBoxLayout]:
        self.HIDE_WHEN_CLOSE = True
        return self._make_close_handle(title)


class IndicatorMixin:
    """Mixin class to instantiate certain methods."""

    evt_indicate = Signal(str)
    evt_indicate_about = Signal(str, str)

    def on_toast(self, title: str, message: str, func: ty.Callable = logger.info) -> None:
        """Show notification."""
        from qtextra.widgets.qt_toast import QtToast

        func(message)
        QtToast(self).show_message(title, message)  # type: ignore[arg-type]

    @staticmethod
    def on_notify_critical(msg: str, func: ty.Callable = logger.critical) -> None:
        """Notify the user of an error."""
        EVENTS.evt_msg_critical.emit(msg)
        func(msg)

    @staticmethod
    def on_notify_error(msg: str, func: ty.Callable = logger.error) -> None:
        """Notify the user of an error."""
        EVENTS.evt_msg_error.emit(msg)
        func(msg)

    @staticmethod
    def on_notify_warning(msg: str, func: ty.Callable = logger.warning) -> None:
        """Notify the user of a warning."""
        EVENTS.evt_msg_warning.emit(msg)
        func(msg)

    @staticmethod
    def on_notify_info(msg: str, func: ty.Callable = logger.info) -> None:
        """Notify the user of an info."""
        EVENTS.evt_msg_info.emit(msg)
        func(msg)

    @staticmethod
    def on_notify_success(msg: str, func: ty.Callable = logger.success) -> None:
        """Notify the user of an success."""
        EVENTS.evt_msg_success.emit(msg)
        func(msg)

    def _indicate_success(self, source: ty.Optional[str] = None) -> None:
        if source and isinstance(source, str):
            self.evt_indicate_about.emit("success", source)
        else:
            self.evt_indicate.emit("success")

    def _indicate_success_any(self, *_args: ty.Any, **_kwargs: ty.Any) -> None:
        self._indicate_success()

    def _indicate_failure(self, source: ty.Optional[str] = None) -> None:
        if source:
            self.evt_indicate_about.emit("warning", source)
        else:
            self.evt_indicate.emit("warning")



"""Base config."""

import glob
import os
import typing as ty
from pathlib import Path

from koyo.json import read_json_data, write_json_data
from loguru import logger
from natsort import natsorted
from qtpy.QtCore import QObject, Signal

from qtextra.utils.appdirs import USER_CONFIG_DIR


class ConfigBase(QObject):
    """Configuration file base."""

    DEFAULT_CONFIG_NAME: str = "config.json"
    DEFAULT_CONFIG_GROUPS: ty.Tuple[str, ...] = ()

    _is_saved: bool = False

    evt_config_loaded = Signal()

    @property
    def output_path(self) -> str:
        """Get default output path."""
        return os.path.join(USER_CONFIG_DIR, self.DEFAULT_CONFIG_NAME)

    @property
    def saved(self) -> bool:
        """Returns indication whether dataset has been recently saved."""
        return self._is_saved

    @saved.setter
    def saved(self, value: bool):
        self._is_saved = value

    def save_config(self, path: ty.Optional[str] = None):
        """Export configuration file to JSON file."""
        if path is None:
            path = self.output_path
        config = {}
        config = self._get_config_parameters(config)

        # write to json file
        write_json_data(path, config)
        self.saved = True
        logger.debug(f"Saved themes to `{path}`")

    @staticmethod
    def _get_config_parameters(config: ty.Dict) -> ty.Dict:
        """Get configuration parameters."""
        return config

    def load_config(self, path: ty.Optional[str] = None, check_type: bool = True):
        """Load configuration from JSON file."""
        from json.decoder import JSONDecodeError

        def _set_values(_group_config):
            for key, new_value in _group_config.items():
                if hasattr(self, key):
                    if check_type:
                        current_value = getattr(self, key)
                        result, _alternative_value = self._check_type(key, current_value, new_value)
                        if not result:
                            logger.warning(
                                f"Could not set `{key}` as the types were not similar enough to ensure compliance."
                                f"\nCurrent value={current_value} ({type(current_value)});"
                                f" New value={new_value} ({type(current_value)})"
                            )
                            if _alternative_value is not None:
                                setattr(self, key, _alternative_value)
                            continue
                    setattr(self, key, new_value)

        if path is None:
            path = self.output_path
        try:
            config = read_json_data(path)
        except FileNotFoundError:
            logger.warning(f"Configuration file does not exist : {path}")
            return
        except JSONDecodeError:
            logger.warning("Could not decode configuration file.")
            return

        if not isinstance(config, dict):
            logger.error("Configuration file should be a dictionary with key:value pairs")
            return

        # iterate over the major groups of settings
        for config_group_title in self.DEFAULT_CONFIG_GROUPS:
            _config_group = config.get(config_group_title, {})
            if not isinstance(_config_group, dict):
                continue
            # check if its dictionary of dictionaries
            if all(isinstance(group_config, dict) for group_config in _config_group.values()):
                for _, group_config in _config_group.items():
                    _set_values(group_config)
            else:
                _set_values(_config_group)

            logger.debug(f"Loaded `{config_group_title}` settings")
        self._set_config_parameters(config)
        logger.debug(f"Loaded config file from `{path}`")

        self.evt_config_loaded.emit()

    def _set_config_parameters(self, config: ty.Dict):
        """Set configuration parameters."""

    def _check_type(self, name, current_value, new_value):
        """Check whether type of the value matches that of the currently set value."""
        current_type = type(current_value)
        new_type = type(new_value)

        # simplest case where types match perfectly
        if current_type == new_type:
            if hasattr(self, f"{name}_choices"):
                choices = getattr(self, f"{name}_choices")
                if isinstance(choices, (list, tuple)):
                    if new_value not in choices:
                        return False, current_value
            return True, None
        else:
            if hasattr(self, f"{name}_validator"):
                validator = getattr(self, f"{name}_validator")
                try:
                    new_value = validator(new_value)
                    return True, new_value
                except Exception:
                    pass
        if current_type in [int, float] and new_type in [int, float]:
            return True, None
        if current_type in [list, tuple] and new_type in [list, tuple]:
            return True, None
        if current_value is None:
            return True, None
        return False, None


def _get_previous_configs(
    base_dir: ty.Optional[str] = None, filename: str = "qtextra-config.json"
) -> ty.Dict[str, str]:
    """Return dictionary of version : path of previous configuration files."""
    if base_dir is None:
        base_dir = USER_CONFIG_DIR
    i = 2
    while not base_dir.endswith("qtextra") and i > 0:
        base_dir = os.path.dirname(base_dir)
        i -= 1

    filelist = natsorted(glob.glob(os.path.join(base_dir, "*")), reverse=True)
    file_dict = {}
    for _, filepath in enumerate(filelist):
        ver_dir = os.path.basename(filepath)
        for file in Path(filepath).glob(f"**/{filename}"):
            if not file_dict:
                ver_dir += " (latest)"
            file_dict[ver_dir] = str(file)

    return file_dict



from pprint import pformat

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QCheckBox,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from qtextra.dialogs.sentry.utilities import PACKAGE, get_sample_event
from qtextra.helpers import get_parent
from qtextra.widgets.qt_dialog import QtDialog


class TelemetryOptInDialog(QtDialog):
    """Opt-in widget."""

    def __init__(self, parent=None, with_locals=False) -> None:
        parent = get_parent(parent)
        super().__init__(parent=parent)
        self._mock_initialized = False
        self._no = False
        self._send_locals = False

        self.send_locals.setChecked(with_locals)
        self._update_example()
        self.resize(720, 740)

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QVBoxLayout:
        """Dialog to provide feedback."""
        btn_box = QDialogButtonBox()
        btn_box.addButton(f"Yes, send my bug reports to {PACKAGE}", QDialogButtonBox.AcceptRole)
        no = btn_box.addButton("No, I'd prefer not to send bug reports", QDialogButtonBox.RejectRole)
        no.clicked.connect(self._set_no)

        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)

        info = QLabel(
            f"""<h3>{PACKAGE} error reporting</h3>
            <br><br>
            Would you like to help us improve '{PACKAGE}' by automatically sending
            bug reports when an error is detected in '{PACKAGE}'?
            <br><br>
            Reports are collected via <a href="https://sentry.io/">Sentry.io</a>
            <br><br>
            Here is an example error log that would be sent from your system:
            """
        )
        info.setWordWrap(True)
        info.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)
        info.setOpenExternalLinks(True)

        self.txt = QTextEdit()
        self.txt.setReadOnly(True)

        self.send_locals = QCheckBox("Include local variables")
        self.send_locals.stateChanged.connect(self._update_example)
        self.send_locals.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        _lbl = QLabel(
            "<small><b>greatly</b> improves interpretability of errors, but may "
            "leak personal identifiable information like file paths</small>"
        )
        _lbl.setWordWrap(True)
        _lbl.setStyleSheet("color: #999;")

        _lbl2 = QLabel("<small>You may change your settings at any time in the Help menu.</small>")
        _lbl2.setStyleSheet("color: #999;")
        _lbl2.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)

        w = QWidget()
        layout = QHBoxLayout()
        layout.addWidget(self.send_locals)
        layout.addWidget(_lbl)
        layout.setContentsMargins(0, 0, 0, 0)
        w.setLayout(layout)

        main_layout = QVBoxLayout()
        main_layout.addWidget(info)
        main_layout.addWidget(self.txt)
        main_layout.addWidget(w)
        main_layout.addWidget(btn_box)
        main_layout.addWidget(_lbl2)
        return main_layout

    def _set_no(self):
        self._no = True

    def _update_example(self):
        self._send_locals = self.send_locals.isChecked()
        event = get_sample_event(include_local_variables=self._send_locals)

        try:
            import yaml

            estring = yaml.safe_dump(event, indent=4, width=120)
        except Exception:
            estring = pformat(event, indent=2, width=120)
        self.txt.setText(estring)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import apply_style, qapplication

    app = qapplication(1)
    dlg = TelemetryOptInDialog(None)
    dlg.setMinimumSize(1200, 500)
    apply_style(dlg)

    dlg.show()
    sys.exit(app.exec_())



"""Message widget."""

from __future__ import annotations

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QVBoxLayout, QWidget

from qtextra import helpers as hp
from qtextra.widgets.qt_label import QtWelcomeLabel


class MessageWidget(QWidget):
    """Widget that displays a message at the center of the page."""

    def __init__(self, message: str, parent=None, icon: str = ""):
        super().__init__(parent)

        self.icon = hp.make_qta_label(self, icon)
        self.icon.setVisible(icon != "")

        self.label = QtWelcomeLabel(message)
        self.label.setWordWrap(True)
        self.label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        layout = QVBoxLayout(self)
        layout.addStretch(True)
        layout.addWidget(self.icon)
        layout.addWidget(self.label)
        layout.addStretch(True)

    @property
    def message(self) -> str:
        """Return current message."""
        return self.label.text()

    @message.setter
    def message(self, value: str):
        self.label.setText(value)



from __future__ import annotations

import typing as ty

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QHBoxLayout, QLayout, QScrollArea, QSizePolicy, QVBoxLayout, QWidget

import qtextra.helpers as hp


class QtScrollableLayoutWidget(QScrollArea):
    """Scrollable horizontal layout."""

    MIN_HEIGHT: int = 30

    def __init__(self, parent: ty.Optional[QWidget] = None):
        super().__init__(parent)

        self._widget = QWidget()
        self._main_layout = self._make_layout()
        self.setWidgetResizable(True)
        self.setWidget(self._widget)
        self.setContentsMargins(0, 0, 0, 0)

    def _make_layout(self) -> QHBoxLayout | QVBoxLayout:
        raise NotImplementedError("Must implement method")

    def count(self) -> int:
        """Return count of widgets."""
        return self._main_layout.count()

    def get_widget(self, index: int) -> QWidget | None:
        """Get widget at position."""
        item = self._main_layout.itemAt(index)
        if not item:
            return None
        widget = item.widget()
        return widget

    def setSpacing(self, spacing: int) -> None:
        """Set spacing."""
        self._main_layout.setSpacing(spacing)

    def addWidget(self, widget: QWidget, **kwargs: ty.Any) -> None:
        """Add widget."""
        self._main_layout.addWidget(widget, **kwargs)
        self._update_scroll()

    def insertWidget(self, index: int, widget: QWidget, **kwargs: ty.Any) -> None:
        """Insert widget."""
        self._main_layout.insertWidget(index, widget, **kwargs)
        self._update_scroll()

    def addLayout(self, layout: QLayout, **kwargs: ty.Any) -> None:
        """Add layout."""
        self._main_layout.addLayout(layout, **kwargs)
        self._update_scroll()

    def insertLayout(self, index: int, layout: QLayout, **kwargs: ty.Any) -> None:
        """Insert layout."""
        self._main_layout.insertLayout(index, layout, **kwargs)
        self._update_scroll()

    def removeWidgetOrLayout(self, index: int) -> None:
        """Remove widget or layout based on index position."""
        widget = self.get_widget(index)
        if widget:
            self._main_layout.removeWidget(widget)
            widget.deleteLater()
            self._update_scroll()

    def minimum_height_for_widgets(self) -> int:
        """Return minimum height for all widgets."""
        height = 0
        for i in range(self.count() - 1):
            widget = self.get_widget(i)
            if widget:
                height = max(height, widget.minimumHeight())
        return height

    def minimum_width_for_widgets(self) -> int:
        """Return minimum height for all widgets."""
        height = 0
        for i in range(self.count() - 1):
            widget = self.get_widget(i)
            if widget:
                height = max(height, widget.minimumWidth())
        return height

    def set_min_height(self, min_height: int) -> None:
        """Set minimum height."""
        self.MIN_HEIGHT = min_height
        self._update_scroll()

    def _update_scroll(self) -> None:
        """Update scroll."""


class QtScrollableHLayoutWidget(QtScrollableLayoutWidget):
    """Scrollable horizontal layout."""

    MIN_HEIGHT: int = 30

    def __init__(self, parent: ty.Optional[QWidget] = None):
        super().__init__(parent)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

    def _make_layout(self) -> QHBoxLayout:
        return hp.make_h_layout(stretch_after=True, spacing=1, margin=1, parent=self._widget)

    def set_min_height(self, min_height: int) -> None:
        """Set minimum height."""
        self.MIN_HEIGHT = min_height
        self._update_scroll()

    def _update_scroll(self) -> None:
        """Update scroll."""
        height = self.MIN_HEIGHT
        if self.count() > 1:
            height = max(height, self.minimum_height_for_widgets())
        if self.maximumHeight() != height:
            self.setMaximumHeight(height + 8)


class QtScrollableVLayoutWidget(QtScrollableLayoutWidget):
    """Scrollable vertical layout."""

    MIN_WIDTH: int = 30

    def __init__(self, parent: ty.Optional[QWidget] = None):
        super().__init__(parent)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)

    def _make_layout(self) -> QHBoxLayout:
        return hp.make_v_layout(stretch_after=True, spacing=1, margin=1, parent=self._widget)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QPushButton

    from qtextra.utils.dev import qframe

    random_text = ["Short", "Longer", "Different text", "More text", "Even longer button text"] * 4

    app, frame, ha = qframe(horz=False)
    layout = QtScrollableHLayoutWidget(None)
    ha.addWidget(layout)
    for text in random_text:
        layout.addWidget(QPushButton(text))

    layout = QtScrollableVLayoutWidget(None)
    ha.addWidget(layout)
    for text in random_text:
        layout.addWidget(QPushButton(text))
    frame.show()
    sys.exit(app.exec_())



"""Table view."""

from __future__ import annotations

import operator
import typing as ty
from contextlib import contextmanager, suppress
from enum import Enum

import numpy as np
from loguru import logger
from natsort.natsort import index_natsorted, order_by_index
from qtpy.QtCore import (  # type: ignore[attr-defined]
    QAbstractTableModel,
    QModelIndex,
    QSize,
    QSortFilterProxyModel,
    Qt,
    Signal,
    Slot,
)
from qtpy.QtGui import QBrush, QColor, QKeyEvent, QPainter, QTextDocument
from qtpy.QtWidgets import QAbstractItemView, QHeaderView, QStyledItemDelegate, QStyleOptionViewItem, QTableView
from superqt.utils import ensure_main_thread

from qtextra.config import THEMES
from qtextra.helpers import make_qta_icon, qt_signals_blocked
from qtextra.utils.color import get_text_color
from qtextra.utils.table_config import TableConfig
from qtextra.utils.utilities import connect

TEXT_COLOR: str = "#000000"
LINK_COLOR: str = "#0000FF"

__all__ = (
    "MultiColumnMultiValueProxyModel",
    "MultiColumnSingleValueProxyModel",
    "QtCheckableItemModel",
    "QtCheckableTableView",
    "SingleColumnMultiValueProxyModel",
    "TableConfig",
)


class WrapTextDelegate(QStyledItemDelegate):
    """Wrap text delegate."""

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index) -> None:
        """Paint."""
        text = index.model().data(index)
        document = QTextDocument()
        document.setTextWidth(option.rect.width())  # Set text width to cell width
        document.setDefaultFont(option.font)
        document.setHtml(str(text))

        painter.save()
        painter.translate(option.rect.topLeft())
        document.drawContents(painter)
        painter.restore()

    def sizeHint(self, option: QStyleOptionViewItem, index) -> QSize:
        """Size hint."""
        text = index.model().data(index)
        document = QTextDocument()
        document.setTextWidth(option.rect.width())  # Set text width to option's width
        document.setDefaultFont(option.font)
        document.setHtml(text)
        return QSize(document.idealWidth(), int(document.size().height() / 4.5))


class MultiFilterMode(str, Enum):
    """Multi filter mode."""

    OR = "OR"
    AND = "AND"


class FilterProxyModelBase(QSortFilterProxyModel):
    """Base class for filters."""

    compare_func: ty.Callable[[ty.Iterable[bool]], bool]
    is_multi_or: bool
    sourceModel: ty.Callable[[], QtCheckableItemModel]

    evt_filtered = Signal()

    def __init__(self, *args: ty.Any, mode: MultiFilterMode = MultiFilterMode.AND, **kwargs: ty.Any):
        QSortFilterProxyModel.__init__(self, *args, **kwargs)
        self._multi_filter_mode = mode
        self.multi_filter_mode = mode

    @property
    def multi_filter_mode(self) -> MultiFilterMode:
        """Return multi filter mode."""
        return self._multi_filter_mode

    @multi_filter_mode.setter
    def multi_filter_mode(self, value: MultiFilterMode) -> None:
        self._multi_filter_mode = value
        self.is_multi_or = self.multi_filter_mode == MultiFilterMode.OR
        self.compare_func = any if self.is_multi_or else all

    def sort(self, column: int, order: Qt.SortOrder | None = None) -> None:
        """Sort table."""
        if self.sourceModel().no_sort_columns and column in self.sourceModel().no_sort_columns:
            return
        super().sort(column, order)

    def setFilterByColumn(self, *args: ty.Any) -> None:
        """Set filter by column."""
        raise NotImplementedError("Must implement method")

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        """Filter rows."""
        raise NotImplementedError("Must implement method")

    def find_visible_rows(self) -> tuple[list[int], list[int]]:
        """Find visible rows."""
        visible_rows = []
        hidden_rows = []
        source_model = self.sourceModel()
        for row in range(source_model.rowCount()):
            source_index = source_model.index(row, 0)
            proxy_index = self.mapFromSource(source_index)
            if proxy_index.isValid():
                visible_rows.append(row)
            else:
                hidden_rows.append(row)
        return visible_rows, hidden_rows


class MultiColumnSingleValueProxyModel(FilterProxyModelBase):
    """Proxy model to filter by."""

    def __init__(self, *args: ty.Any, mode: MultiFilterMode = MultiFilterMode.AND, **kwargs: ty.Any):
        super().__init__(*args, mode=mode, **kwargs)
        self.filters_by_text: dict[int, str] = {}
        self.filters_by_state: dict[int, Qt.CheckState] = {}

    def setFilterByColumn(self, text: str, column: int) -> None:
        """Set filter by column."""
        if not text and column in self.filters_by_text:
            del self.filters_by_text[column]
        self.filters_by_text[column] = str(text).lower()
        self.invalidateFilter()
        self.evt_filtered.emit()

    def setFilterByState(self, value: bool | None, column: int) -> None:
        """Set filter by value."""
        if value is None and column in self.filters_by_state:
            del self.filters_by_state[column]
        else:
            self.filters_by_state[column] = Qt.CheckState.Checked if value else Qt.CheckState.Unchecked
        self.invalidateFilter()
        self.evt_filtered.emit()

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        """Filter rows."""
        if not self.filters_by_text and not self.filters_by_state:
            return True

        results: list[bool] = []
        for column, state in self.filters_by_state.items():
            index = self.sourceModel().index(source_row, column, source_parent)
            if index.isValid():
                value = self.sourceModel().data(index, Qt.ItemDataRole.CheckStateRole)
                results.append(value == state)

        for column, text in self.filters_by_text.items():
            value = ""
            index = self.sourceModel().index(source_row, column, source_parent)
            if index.isValid():
                value = self.sourceModel().data(index, Qt.ItemDataRole.DisplayRole)
                if not value:
                    return True
            results.append(text in value.lower())
        return self.compare_func(results)


class MultiColumnMultiValueProxyModel(FilterProxyModelBase):
    """Proxy model to filter by."""

    def __init__(
        self,
        *args: ty.Any,
        mode: MultiFilterMode = MultiFilterMode.AND,
        column_mode: MultiFilterMode = MultiFilterMode.OR,
        **kwargs: ty.Any,
    ):
        super().__init__(*args, mode=mode, **kwargs)
        self.column_mode = column_mode
        # self.column_compare_func = any if column_mode == MultiFilterMode.OR else all
        self.filters_by_text: dict[int, list[str]] = {}
        self.column_compare_funcs: dict[int, ty.Callable[[ty.Iterable[bool]], bool]] = {}

    def column_compare_func(self, column: int) -> ty.Callable:
        """Return instance of callable for specific column."""
        if column in self.column_compare_funcs:
            return self.column_compare_funcs[column]
        return any if self.column_mode == MultiFilterMode.OR else all

    def setFilterByColumn(self, filters: list[str], column: int, column_mode: MultiFilterMode | None = None) -> None:
        """Set filter by column."""
        if not filters:
            if column in self.filters_by_text:
                del self.filters_by_text[column]
            if column in self.column_compare_funcs:
                del self.column_compare_funcs[column]
        if not isinstance(filters, list):
            filters = [filters]
        if filters:
            self.filters_by_text[column] = [filt.lower() for filt in filters]
            if column_mode:
                self.column_compare_funcs[column] = any if column_mode == MultiFilterMode.OR else all
        self.invalidateFilter()
        self.evt_filtered.emit()

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        """Filter rows."""
        if not self.filters_by_text:
            return True
        results: list[bool] = []
        for column, _texts in self.filters_by_text.items():
            value = ""
            index = self.sourceModel().index(source_row, column, source_parent)
            if index.isValid():
                value = self.sourceModel().data(index, Qt.ItemDataRole.DisplayRole)
                if not value:
                    return True
            results.append(self.column_compare_func(column)(text in value.lower() for text in _texts))
        return self.compare_func(results)


class SingleColumnMultiValueProxyModel(FilterProxyModelBase):
    """Proxy model to filter by."""

    def __init__(self, *args: ty.Any, column: int, mode: MultiFilterMode = MultiFilterMode.AND, **kwargs: ty.Any):
        super().__init__(*args, mode=mode, **kwargs)
        self.column = column
        self.filters: list[str] = []

    def setFilterByColumn(self, filters: list[str], column: int | None = None) -> None:
        """Set filter by column."""
        if column is not None:
            self.column = column
        if not isinstance(filters, list):
            filters = [filters]

        self.filters = [filt.lower() for filt in filters]
        self.invalidateFilter()
        self.evt_filtered.emit()

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        """Filter rows."""
        if not self.filters:
            return True
        # if not source_parent.isValid():
        #     return True

        results: list[bool] = []
        for text in self.filters:
            value = ""
            index = self.sourceModel().index(source_row, self.column, source_parent)
            if index.isValid():
                value = self.sourceModel().data(index, Qt.ItemDataRole.DisplayRole)
                if not value:
                    return True
            results.append(text in value.lower())
        return self.compare_func(results)


class QtCheckableItemModel(QAbstractTableModel):
    """Checkable item model."""

    evt_checked = Signal(int, bool)
    evt_value_checked = Signal(int, int, bool)

    table_proxy: MultiColumnSingleValueProxyModel | SingleColumnMultiValueProxyModel | None = None

    def __init__(
        self,
        parent: QtCheckableTableView,
        data: list[list],
        header: list[str] | None = None,
        no_sort_columns: list[int] | None = None,
        hidden_columns: list[int] | None = None,
        color_columns: list[int] | None = None,
        html_columns: list[int] | None = None,
        icon_columns: list[int] | None = None,
        checkable_columns: list[int] | None = None,
        text_alignment: Qt.AlignmentFlag | None = None,
    ):
        QAbstractTableModel.__init__(self, parent)
        self._table: list[list[ty.Any]] = data
        self.state = None
        self.original_index = list(range(len(self._table)))
        self.header = header or []
        self.no_sort_columns = no_sort_columns or []
        self.hidden_columns = hidden_columns or []
        self.color_columns = color_columns or []
        self.html_columns = html_columns or []
        self.icon_columns = icon_columns or []
        self.checkable_columns = checkable_columns or []
        self.text_alignment = text_alignment or Qt.AlignmentFlag.AlignCenter

    def flags(self, index):
        """Return flags."""
        fl = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        column = index.column()
        if column == 0 or column in self.checkable_columns:
            fl |= Qt.ItemFlag.ItemIsUserCheckable
        else:
            fl |= Qt.ItemFlag.ItemIsEditable
        return fl

    @property
    def n_checked(self) -> int:
        """Return of checked elements."""
        return [row[0] for row in self._table].count(True)

    @property
    def n_unchecked(self) -> int:
        """Return count of unchecked elements."""
        return [row[0] for row in self._table].count(False)

    def rowCount(self, parent: QModelIndex | None = None, **kwargs: ty.Any) -> int:
        """Return number of rows."""
        return len(self._table) if self._table else 0

    def columnCount(self, parent: QModelIndex | None = None, **kwargs: ty.Any) -> int:
        """Return number of columns."""
        return len(self._table[0]) if self._table else 0

    def removeRow(self, row: int, parent: QModelIndex | None = None) -> bool:
        """Remove row."""
        if parent is None:
            parent = QModelIndex()
        self.beginRemoveRows(parent, row, row)
        self._table.pop(row)
        self.original_index.pop(row)
        self.endRemoveRows()
        return True

    def data(self, index: QModelIndex, role: Qt.ItemDataRole | None = None) -> ty.Any:
        """Parse data."""
        if not index.isValid():
            return None

        row = index.row()
        column = index.column()

        # check the background color
        if role == Qt.ItemDataRole.BackgroundRole:
            if column in self.color_columns:
                color = self._table[row][column]
                if isinstance(color, str) and "#" in color:
                    return QBrush(QColor(color))
                elif isinstance(color, np.ndarray):
                    return QBrush(QColor(*(255 * color).astype("int")))
                if isinstance(color, QColor):
                    return QBrush(color)
            return QBrush()
        # check text color
        elif role == Qt.ItemDataRole.ForegroundRole:
            if column in self.color_columns:
                bg_color = self._table[row][column]
                if isinstance(bg_color, str) and "#" in bg_color:
                    return QBrush(get_text_color(QColor(bg_color)))
            # let's use slightly different color html
            if column in self.html_columns:
                return QBrush(QColor(LINK_COLOR))
            return QBrush(QColor(TEXT_COLOR))
        # check value
        elif role == Qt.ItemDataRole.DisplayRole:
            if column not in self.icon_columns and column not in self.checkable_columns:
                value = self._table[row][column]
                return value
        # check the alignment role
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return self.text_alignment
        # check state
        elif role == Qt.ItemDataRole.CheckStateRole and column in self.checkable_columns:
            return Qt.CheckState.Checked if self._table[row][column] else Qt.CheckState.Unchecked
        # icon state
        elif role == Qt.ItemDataRole.DecorationRole:
            if column in self.icon_columns:
                value = self._table[row][column]
                return make_qta_icon(value)

    def headerData(self, col: int, orientation: Qt.Orientation, role: Qt.ItemDataRole | None = None) -> str | None:
        """Get header data."""
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self.header[col]
        return None

    def sort(self, column: int, order: Qt.SortOrder | None = None) -> None:
        """Sort table."""
        if self.no_sort_columns and column in self.no_sort_columns:
            return

        # emit signal about upcoming change
        self.layoutAboutToBeChanged.emit()

        # get sort index
        new_index = index_natsorted(self._table, key=operator.itemgetter(column))

        # sort
        self._table = order_by_index(self._table, new_index)
        self.original_index = order_by_index(self.original_index, new_index)

        if order == Qt.SortOrder.DescendingOrder:
            self._table.reverse()
            self.original_index.reverse()

        # indicate that change to data has been made
        self.layoutChanged.emit()

    def setData(self, index: QModelIndex, value: ty.Any, role=Qt.ItemDataRole.EditRole) -> bool:
        """Set data in the model."""
        row = index.row()
        column = index.column()

        if role == Qt.ItemDataRole.CheckStateRole:
            old_value = index.data()
            # value = not old_value
            # change = True
            value = bool(value)
            change = old_value != value
        else:
            old_value = index.data()
            if isinstance(old_value, np.ndarray):
                change = np.any(old_value != value)
            else:
                change = old_value != value

        self._table[row][column] = value
        if change:
            self.dataChanged.emit(index, index)
            if column == 0:
                self.evt_checked.emit(row, value)
            elif column in self.checkable_columns:
                self.evt_value_checked.emit(row, column, value)
            return True
        return False

    def update_value(self, row: int, column: int, value: ty.Any, role=Qt.ItemDataRole.EditRole) -> None:
        """Update value."""
        index = self.createIndex(row, column)

        # setup role
        if column == 0 or column in self.checkable_columns:
            role = Qt.ItemDataRole.CheckStateRole

        if index.isValid():
            self.setData(index, value, role=role)

    def update_values(self, row, column_value) -> None:
        """Update values."""
        for column, value in column_value.items():
            self.update_value(row, column, value)

    def update_row(self, row: int, value: ty.Any) -> None:
        """Update row."""
        if len(value) != len(self.header):
            raise ValueError("Cannot update row as length of the values does not match header length")

        for column in range(len(self.header)):
            index = self.createIndex(row, column)
            if index.isValid():
                self.setData(index, value[column])

    def update_column(self, col, values, match_to_sort: bool = True) -> None:
        """Update column."""
        if col > self.columnCount():
            raise ValueError("Cannot update column as its outside of the boundaries")

        if len(values) > self.rowCount():
            raise ValueError("Cannot update column as the length of the values does not match the number of rows")

        for row, value in enumerate(values):
            if match_to_sort:
                row = self.get_sort_index(row)
            index = self.createIndex(row, col)
            if index.isValid():
                self.setData(index, value)

    def toggle_all_rows(self) -> None:
        """Check all rows in the table."""
        if self.state is None:
            self.state = self.n_checked == self.rowCount()
        self.state = not self.state
        for row, __ in enumerate(self._table):
            if self.table_proxy and not self.table_proxy.filterAcceptsRow(row, QModelIndex()):
                continue
            self._table[row][0] = self.state
            index = self.createIndex(row, 0)
            self.dataChanged.emit(index, index)
        self.evt_checked.emit(-1, self.state)

    def check_all_rows(self) -> None:
        """Check all rows in the table."""
        for row, __ in enumerate(self._table):
            if self.table_proxy and not self.table_proxy.filterAcceptsRow(row, QModelIndex()):
                continue
            self._table[row][0] = True
            index = self.createIndex(row, 0)
            self.dataChanged.emit(index, index)
        self.evt_checked.emit(-1, True)

    def uncheck_all_rows(self) -> None:
        """Uncheck all rows."""
        for row, __ in enumerate(self._table):
            if self.table_proxy and not self.table_proxy.filterAcceptsRow(row, QModelIndex()):
                continue
            self._table[row][0] = False
            index = self.createIndex(row, 0)
            self.dataChanged.emit(index, index)
        self.evt_checked.emit(-1, False)

    def get_all_checked(self) -> list[int]:
        """Get all checked items."""
        return self._get_all_state(True)

    def get_all_unchecked(self) -> list[int]:
        """Get all unchecked items."""
        return self._get_all_state(False)

    def _get_all_state(self, state: bool) -> list[int]:
        """Get all checked items."""
        checked = []
        for i, row in enumerate(self._table):
            if row[0] is state:
                checked.append(i)
        return checked

    def roleNames(self):
        """Return role names."""
        roles = QAbstractTableModel.roleNames(self)
        roles[hash("Checked")] = Qt.ItemDataRole.CheckStateRole
        return roles

    def reset(self):
        """Reset model."""
        self.beginResetModel()
        self.endResetModel()

    def get_initial_index(self, row: int) -> int:
        """Get the index of the initial array, regardless of whether it was sorted."""
        return self.original_index[row]

    def get_initial_indices(self, index: list) -> list:
        """Get list of all initial indices."""
        return [self.get_initial_index(row) for row in index]

    def get_sort_index(self, row: int) -> int:
        """Get the index inside the sorted array as matched from the not-sorted array."""
        return self.original_index.index(row)

    def get_sort_indices(self, index: list) -> list:
        """Get list of all sort indices."""
        return [self.get_sort_index(row) for row in index]

    def get_data(self) -> list[list]:
        """Get data from model."""
        return self._table

    def get_row_id(self, col_id: int, value: str) -> int:
        """Find value index."""
        for row_id, row in enumerate(self._table):
            if row[col_id] == value:
                return row_id
        return -1

    def get_row_id_for_values(self, *column_and_values: ty.Tuple[int, ty.Union[str, int, float]]) -> int:
        """Find value index."""
        for row_id, row in enumerate(self._table):
            for col_id, value in column_and_values:
                if row[col_id] != value:
                    break
                if all(row[col_id] == value for col_id, value in column_and_values):
                    return row_id
        return -1

    @ensure_main_thread
    def add_data(self, data: list) -> None:
        """Add data."""
        self._table.extend(data)
        self.original_index = list(range(len(self._table)))
        # indicate that change to data has been made
        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    @ensure_main_thread
    def reset_data(self) -> None:
        """Reset data."""
        self._table.clear()
        self.original_index.clear()
        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    def data_changed(self) -> None:
        """Emit an event when there has been change to the model."""
        self.layoutChanged.emit()


class QtCheckableTableView(QTableView):
    """Checkbox table."""

    model: ty.Callable[[], QtCheckableItemModel]

    # events
    # triggered whenever item is checked/unchecked. It returns the index and check state when its triggered.
    # It behaves slightly differently when user clicks on the header and -1 is emitted rather than actual index
    evt_checked = Signal(int, bool)
    evt_value_checked = Signal(int, int, bool)
    # keyboard event
    evt_keypress = Signal(QKeyEvent)
    # value changed
    evt_changed = Signal()
    evt_double_clicked = Signal(int)
    evt_reset = Signal()

    def __init__(
        self,
        *args: ty.Any,
        config: TableConfig | None = None,
        enable_all_check: bool = True,
        double_click_to_check: bool = False,
        sortable: bool = True,
        checkable: bool = False,
        drag: bool = False,
        selection: QAbstractItemView.SelectionMode = QAbstractItemView.SelectionMode.ExtendedSelection,
        **kwargs: ty.Any,
    ):
        super().__init__(*args, **kwargs)

        # setup config
        self._config = config
        self._header_columns = None
        self._is_init = False
        self.enable_all_check = enable_all_check
        self.checkable = checkable

        self._double_click_to_check = double_click_to_check
        self._sortable = sortable
        self._drag = drag
        self._selection = selection

        # register events
        self.doubleClicked.connect(lambda v: self.evt_double_clicked.emit(v.row()))
        self.clicked.connect(self.on_table_clicked)
        if self._sortable:
            self.header.sectionClicked.connect(self.sortByColumn)
        if self._double_click_to_check:
            self.doubleClicked.connect(self._on_check_row)
        if isinstance(self._config, TableConfig):
            self.init_from_config()
        connect(THEMES.evt_theme_changed, self._update_color_theme, state=True)
        self._update_color_theme()
        self.setCornerButtonEnabled(False)

    def closeEvent(self, event) -> None:
        """Close event."""
        connect(THEMES.evt_theme_changed, self._update_color_theme, state=False)
        return super().closeEvent(event)

    def activate_word_wrap(self) -> None:
        """Set word wrap."""
        self.setWordWrap(True)
        self.setItemDelegate(WrapTextDelegate())  # Use the custom delegate

    def resize_wrapped_rows(self) -> None:
        """Resize wrapped rows."""
        # This is a simplistic approach to dynamically calculate row height.
        # You might need a more sophisticated calculation based on your content.
        for row in range(self.model().rowCount()):
            new_height = 0
            for column in range(self.model().columnCount()):
                index = self.model().index(row, column)
                height = self.itemDelegate().sizeHint(QStyleOptionViewItem(), index).height()
                if height > new_height:
                    new_height = height
            self.setRowHeight(row, new_height)

    @Slot()
    def _update_color_theme(self):
        """Update global color theme."""
        global TEXT_COLOR, LINK_COLOR
        TEXT_COLOR = THEMES.get_theme_color()
        LINK_COLOR = THEMES.get_hex_color("highlight")
        with suppress(RuntimeError):
            self.update(QModelIndex())

    def _on_check_row(self, evt):
        """Event triggers check/uncheck of row."""
        row_id = evt.row()
        self.update_value(row_id, 0, not self.is_checked(row_id))

    @property
    def n_rows(self) -> int:
        """Return the number of rows in the table."""
        return self.row_count()

    @property
    def n_cols(self) -> int:
        """Return the number of columns in the table."""
        return self.column_count()

    def column_count(self) -> int:
        """Return the number columns."""
        return self.model().columnCount(self) if self.model() else 0

    def row_count(self) -> int:
        """Return the number of rows."""
        return self.model().rowCount(self) if self.model() else 0

    def row_visible_count(self) -> int:
        """Return the number of rows that are actually visible."""
        model = self.model()
        if model:
            if hasattr(model, "table_proxy"):
                return model.table_proxy.rowCount()
            return model.rowCount()
        return 0

    @property
    def header(self):
        """Return header."""
        return self.horizontalHeader()

    @property
    def index(self) -> QHeaderView:
        """Return index."""
        return self.verticalHeader()

    def model(self) -> QtCheckableItemModel:
        """Return instance of model."""
        model: QtCheckableItemModel = super().model()
        if hasattr(model, "sourceModel"):
            return model.sourceModel()
        return model

    def is_proxy(self) -> bool:
        """Return True if model is a proxy model."""
        return isinstance(super().model(), MultiColumnSingleValueProxyModel)

    def proxy_or_model(self) -> QtCheckableItemModel | MultiColumnSingleValueProxyModel:
        """Return instance of model."""
        model: QtCheckableItemModel = super().model()
        if isinstance(model, MultiColumnSingleValueProxyModel):
            return model
        return model

    def on_table_clicked(self, index: ty.Optional[QModelIndex] = None) -> None:
        """Imitate row selection."""
        if index is None:
            index = QModelIndex()
        if not index.isValid():
            return
        row = index.row()
        self.selectRow(row)

    def init(self) -> None:
        """Initialize table to ensure correct visuals."""
        sizing = {
            "stretch": QHeaderView.ResizeMode.Stretch,
            "fixed": QHeaderView.ResizeMode.Fixed,
            "contents": QHeaderView.ResizeMode.ResizeToContents,
        }
        # Get hook for the header
        config: TableConfig | None = self._config
        n_cols = self.column_count()
        header = self.header
        # 25 px is optimal size for checkbox
        header.setMinimumSectionSize(25)
        for column_id in range(n_cols):
            if config:
                sizing_ = config.get(column_id)["sizing"]
                mode = sizing.get(
                    sizing_, QHeaderView.ResizeMode.Stretch if column_id else QHeaderView.ResizeMode.Fixed
                )
            else:
                # The first column should always be a QCheckbox
                mode = QHeaderView.ResizeMode.Fixed if column_id == 0 else QHeaderView.ResizeMode.Stretch
            header.setSectionResizeMode(column_id, mode)
            if config and mode == QHeaderView.ResizeMode.Fixed:
                header.resizeSection(column_id, config.get_width(column_id))

        # set column width for the first column (checkbox)
        self.setColumnWidth(0, 25)

        # disable editing
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        # enable sorting
        self.setSortingEnabled(self._sortable)
        # disable drag
        self.setDragEnabled(self._drag)
        if self._drag:
            self.setDropIndicatorShown(True)
        # set selection mode
        self.setSelectionMode(self._selection)

        # hide columns
        model = self.model()
        for n_col in model.hidden_columns:
            self.setColumnHidden(n_col, True)

        self._is_init = True
        model.data_changed()

    def set_column_resize_mode(self, index: int, mode: QHeaderView.ResizeMode = QHeaderView.ResizeMode.Stretch):
        """Set column resize mode."""
        if self._is_init:
            self.header.setSectionResizeMode(index, mode)

    def on_check(self, row: int, value: bool) -> None:
        """Check."""
        self.evt_checked.emit(row, value)

    def init_from_config(self) -> None:
        """Initialize based on config."""
        self.set_data(
            [],
            self._config.header,
            self._config.no_sort_columns,
            self._config.hidden_columns,
            icon_columns=self._config.icon_columns,
        )

    def set_model(self, model: QtCheckableItemModel) -> None:
        """Set model."""
        if self._config:
            model.icon_columns = self._config.icon_columns
            model.color_columns = self._config.color_columns
            model.html_columns = self._config.html_columns
            model.no_sort_columns = self._config.no_sort_columns
            model.hidden_columns = self._config.hidden_columns
            model.checkable_columns = self._config.checkable_columns
            # model.text_alignment = {
            #     "left": Qt.AlignmentFlag.AlignLeft,
            #     "center": Qt.AlignmentFlag.AlignCenter,
            #     "right": Qt.AlignmentFlag.AlignRight,
            # }[self._config.text_alignment]
        self.setModel(model)

    def setup_model_from_config(self, config: TableConfig):
        """Setup model from config."""
        self.setup_model(
            config.header,
            config.no_sort_columns,
            config.hidden_columns,
            config.html_columns,
            config.icon_columns,
            config.checkable_columns,
            text_alignment={
                "left": Qt.AlignmentFlag.AlignLeft,
                "center": Qt.AlignmentFlag.AlignCenter,
                "right": Qt.AlignmentFlag.AlignRight,
            }[config.text_alignment],
        )

    def setup_model(
        self,
        header: list[str],
        no_sort_columns: list[int] | None = None,
        hidden_columns: list[int] | None = None,
        html_columns: list[int] | None = None,
        icon_columns: list[int] | None = None,
        checkable_columns: list[int] | None = None,
        text_alignment: Qt.AlignmentFlag | None = None,
    ) -> None:
        """Setup model in the table."""
        self.set_data(
            [],
            header,
            no_sort_columns,
            hidden_columns,
            html_columns,
            icon_columns,
            checkable_columns,
            text_alignment=text_alignment,
        )

    def reset_data(self) -> None:
        """Clear table."""
        model = self.model()
        if hasattr(model, "reset_data"):
            model.reset_data()
        self.evt_reset.emit()

    def set_data(
        self,
        data: list,
        header: list[str],
        no_sort_columns: list[int] | None = None,
        hidden_columns: list[int] | None = None,
        html_columns: list[int] | None = None,
        icon_columns: list[int] | None = None,
        color_columns: list[int] | None = None,
        checkable_columns: list[int] | None = None,
        text_alignment: Qt.AlignmentFlag | None = None,
        checkable: bool | str = "auto",
    ) -> None:
        """Set data."""
        self._header_columns = header
        if checkable == "auto" and header:
            checkable = header[0] == ""  # empty column usually indicates that that the first column is checkable
        else:
            checkable = checkable
        self._validate_data(data, len(header))
        model = QtCheckableItemModel(
            self,
            data=data,
            header=header,
            no_sort_columns=no_sort_columns or [],
            hidden_columns=hidden_columns or [],
            color_columns=color_columns or [],
            html_columns=html_columns or [],
            icon_columns=icon_columns or [],
            checkable_columns=checkable_columns or [],
            text_alignment=text_alignment,
        )
        self.checkable = bool(checkable)
        model.evt_checked.connect(self.evt_checked.emit)
        model.evt_value_checked.connect(self.evt_value_checked.emit)
        self.set_model(model)
        self.init()

    def add_row(self, data: list) -> None:
        """ADd row to the data."""
        self.add_data([data])

    def append_data(self, data: list) -> None:
        """Append data."""
        self.add_data(data)

    def add_data(self, data: list[list]) -> None:
        """Add data."""
        n_items = self.n_rows
        self._validate_data(data)
        self.model().add_data(data)
        if n_items == 0:
            self.init()

    def add_index(self, rows: list[str]):
        """Add vertical index."""
        # header.set
        # for i, row in enumerate(rows):
        #     header.setSectionResizeMode(i, QHeaderView.ResizeMode.ResizeToContents)
        #     header.item

    def add_data_without_set(self, data: list[list]) -> None:
        """Add data."""
        n_items = self.n_rows
        self._validate_data(data)
        self.model().add_data(data)
        if n_items == 0:
            self.init()

    def _validate_data(self, data: list, n_cols: ty.Optional[int] = None) -> None:
        """Validate data."""
        if n_cols is None:
            if self._header_columns is not None:
                n_cols = len(self._header_columns)
            else:
                n_cols = self.n_cols
        for _data in data:
            if len(_data) != n_cols:
                logger.warning("Data is of incorrect size")

    def get_data(self) -> list[list]:
        """Get data from model.

        This returns the native data that is stored in the model.
        """
        data = self.model().get_data()
        return data

    def get_all_checked(self) -> list[int]:
        """Get all checked."""
        return self.model().get_all_checked()

    def get_all_shown(self) -> list[int]:
        """Get all checked."""
        model = self.model()
        proxy = model.table_proxy
        if proxy:
            return proxy.find_visible_rows()[0]
        shown = []
        for index in range(model.rowCount()):
            if not self.isRowHidden(index):
                shown.append(index)
        return shown

    def get_all_unchecked(self) -> list[int]:
        """Get all unchecked.

        Returns
        -------
        list of ints
            List of all rows that are currently unchecked.
        """
        return self.model().get_all_unchecked()

    def toggle_all_rows(self) -> None:
        """Uncheck all values."""
        self.model().toggle_all_rows()

    def check_all_rows(self) -> None:
        """Uncheck all values."""
        self.model().check_all_rows()

    def uncheck_all_rows(self) -> None:
        """Uncheck all values."""
        self.model().uncheck_all_rows()

    def get_initial_index(self, indices: list) -> list[int]:
        """Get initial index."""
        return self.model().get_initial_indices(indices)

    def get_row_id(self, col_id: int, value: ty.Any) -> int:
        """Get the id of a value."""
        return self.model().get_row_id(col_id, value)

    def get_row_id_for_values(self, *column_and_values: ty.Tuple[int, ty.Any]) -> int:
        """Get the id of a value."""
        return self.model().get_row_id_for_values(*column_and_values)

    def get_col_data(self, col_id: int) -> list[ty.Any]:
        """Get data from model."""
        data = self.model().get_data()
        if col_id <= self.n_cols:
            data = [row[col_id] for row in data]
        return data

    def get_row_data(self, row_id: int) -> list:
        """Get data from model."""
        data = self.model().get_data()
        if row_id <= self.n_rows:
            data = data[row_id]
        return data

    def is_checked(self, row_id: int) -> bool:
        """Get check state of value."""
        value = self.get_value(0, row_id)
        if value == "":
            return False
        return bool(value)

    def get_value(self, col_id: int, row_id: int) -> ty.Any:
        """Get data from model."""
        data = self.model().get_data()
        if row_id <= self.n_rows and col_id <= self.n_cols:
            data = data[row_id][col_id]
        return data

    def set_value(self, col_id: int, row_id: int, value: ty.Any) -> None:
        """Set value in the data model."""
        self.model().update_value(row_id, col_id, value)

    def select_row(self, row: int, match_to_sort: bool = True) -> None:
        """Select row."""
        if match_to_sort:
            row = self.model().get_sort_index(row)
        self.selectRow(row)

    def update_value(self, row: int, col: int, value: ty.Any, match_to_sort: bool = True) -> None:
        """Update value in the model."""
        if match_to_sort:
            row = self.model().get_sort_index(row)
        self.model().update_value(row, col, value)

    def remove_row(self, row_id: int) -> None:
        """Remove row from the model."""
        self.model().removeRow(row_id)

    def remove_rows(self, rows: list[int]) -> None:
        """Remove rows from the model."""
        rows = sorted(rows, reverse=True)
        for row in rows:
            self.remove_row(row)

    def update_row(self, row: int, value: list, match_to_sort: bool = True) -> None:
        """Update entire row."""
        if match_to_sort:
            row = self.model().get_sort_index(row)
        self.model().update_row(row, value)

    @contextmanager
    def block_model(self) -> ty.Generator[None, None, None]:
        """Block model signals."""
        with qt_signals_blocked(self.model(), block_signals=True):
            yield
        self.model().reset()

    def update_column(self, col: int, values: list, match_to_sort: bool = True, block_signals: bool = False) -> None:
        """Update entire row."""
        assert len(values) == self.n_rows, (
            f"Tried to set incorrect number of rows. Expected {self.n_rows} - got {len(values)}"
        )
        model = self.model()
        with qt_signals_blocked(model, block_signals=block_signals):
            model.update_column(col, values, match_to_sort)

    def update_values(
        self, row: int, column_value: ty.Dict[int, ty.Union[str, int, float, bool]], match_to_sort: bool = True
    ) -> None:
        """Update multiple columns for a particular row."""
        if match_to_sort:
            row = self.model().get_sort_index(row)
        self.model().update_values(row, column_value)

    def sort_by_column(self, column: int, direction: ty.Union[Qt.SortOrder, str] = "ascending") -> None:
        """Sort table by column."""
        if isinstance(direction, str):
            direction = Qt.SortOrder.AscendingOrder if direction == "ascending" else Qt.SortOrder.DescendingOrder
        self.model().sort(column, direction)

    def find_index_of(self, col_id: int, value: ty.Any) -> int:
        """Find index of value. Return -1 if not found."""
        col_data = self.get_col_data(col_id)
        try:
            return col_data.index(value)
        except ValueError:
            return -1

    def find_indices_of(self, col_id: int, value: ty.Any) -> list[int]:
        """Find index of value. Return -1 if not found."""
        col_data = self.get_col_data(col_id)
        indices = [i for i, x in enumerate(col_data) if x == value]
        return indices

    def find_index_of_value_with_indices(self, col_id: int, value: ty.Any, indices: list[int]) -> int:
        """Find index of value. Return -1 if not found."""
        col_data = self.get_col_data(col_id)
        for i in indices:
            if col_data[i] == value:
                return i
        return -1

    def sortByColumn(self, index: int) -> None:
        """Override method."""
        if index == 0 and self.checkable:
            self.header.setSortIndicatorShown(False)
            if self.enable_all_check:
                self.model().toggle_all_rows()
            return
        else:
            self.header.setSortIndicatorShown(True)
        order = self.horizontalHeader().sortIndicatorOrder()
        return QTableView.sortByColumn(self, index, order)

    def keyPressEvent(self, event):
        """Process key event press."""
        super().keyPressEvent(event)
        row = self.currentIndex().row()
        self.selectRow(row)
        event.row = lambda: row  # make row retrieval a function so its compatible with other methods
        self.evt_keypress.emit(event)

    #         # take into account change of order
    #         idx = self.model().get_initial_index(row) if row >= 0 else -1
    #         self.keyPressSignal.emit(idx)

    def create_index(self, row: int = 0, column: int = 0) -> QModelIndex:
        """Create index."""
        return self.model().createIndex(row, column)
